from operator import __add__, __sub__

from collections import namedtuple

import numpy

from .cfdatetime import elements
from .cfdatetime import dt as cf_dt
from .functions  import inspect as cf_inspect
from .units      import Units

from .data.data import Data

from .decorators import _deprecated_kwarg_check


# Define some useful units
_calendar_years  = Units('calendar_years')
_calendar_months = Units('calendar_months')
_days            = Units('days')
_hours           = Units('hours')
_minutes         = Units('minutes')
_seconds         = Units('seconds')

# Define some useful constants
_one_year   = Data(1, 'calendar_years')
_one_day    = Data(1, 'day')
_one_hour   = Data(1, 'hour')
_one_minute = Data(1, 'minute')

# Default month lengths in days
_default_month_lengths = [31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

_default_calendar = 'gregorian'

Offset = namedtuple('offset', ('year', 'month', 'day', 'hour',
                               'minute', 'second', 'microsecond'))

_relational_methods = ('__eq__', '__ne__',
                       '__lt__', '__le__',
                       '__gt__', '__ge__')


class TimeDuration:
    '''A duration of time.

    The duration of time is a number of either calendar years,
    calender months, days, hours, minutes or seconds.

    A calendar year (or month) is an arbitrary year (or month) in an
    arbitrary calendar. A calendar is as part of the time duration,
    but will be taken from the context in which the time duration
    instance is being used. For example, a calendar may be specified
    when creating time intervals (see below for examples).

    A default offset is specified that may be used by some
    applications to temporally position the time duration. For
    example, setting ``cf.TimeDuration(1, 'calendar_month', day=16,
    hour=12)`` will define a duration of one calendar month which, by
    default, starts at 12:00 on the 16th of the month. Note that the
    offset

    **Changing the units**

    The time duration's units may be changed in place by assigning
    equivalent units to the `~cf.TimeDuration.Units` attribute:

    >>> t = cf.TimeDuration(1, 'day')
    >>> t
    <CF TimeDuration: P1D (Y-M-D 00:00:00)>
    >>> t.Units = 's'
    >>> t
    <CF TimeDuration: PT86400.0S (Y-M-D 00:00:00)>
    >>> t.Units = cf.Units('minutes')
    >>> t
    <CF TimeDuration: PT1440.0M (Y-M-D 00:00:00)>
    >>> t.Units = 'calendar_months'
    ValueError: Can't set units (currently <Units: minutes>) to non-equivalent units <Units: calendar_months>

    **Creating time intervals**

    A time interval of exactly the time duration, starting or ending
    at a particular date-time, may be produced with the `interval`
    method. If elements of the start or end date-time are not
    specified, then default values are taken from the `!year`,
    `!month`, `!day`, `!hour`, `!minute`, and `!second` attributes of
    the time duration instance:

    >>> t = cf.TimeDuration(6, 'calendar_months')
    >>> t
    <CF TimeDuration: P6M (Y-M-01 00:00:00)>
    >>> t.interval(cf.dt(1999, 12))
    (cftime.DatetimeGregorian(1999-12-01 00:00:00),
     cftime.DatetimeGregorian(2000-06-01 00:00:00))
    >>> t = cf.TimeDuration(5, 'days', hour=6)
    >>> t
    <CF TimeDuration: P5D (Y-M-D 06:00:00)>
    >>> t.interval(cf.dt(2004, 3, 2), end=True)
    (cftime.DatetimeGregorian(2004-02-26 00:00:00),
     cftime.DatetimeGregorian(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='noleap'), end=True)
    (cftime.DatetimeNoLeap(2004-02-25 00:00:00),
     cftime.DatetimeNoLeap(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True)
    (cftime.Datetime360Day(2004-02-27 00:00:00),
     cftime.Datetime360Day(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True,
    ...     iso='start and duration')
    '2004-02-27 00:00:00/P5D'

    **Comparison operations**

    Comparison operations are defined for `cf.TimeDuration` objects,
    `cf.Data` objects, `numpy` arrays and numbers:

    >>> cf.TimeDuration(2, 'calendar_years') > cf.TimeDuration(
    ...     1, 'calendar_years')
    True
    >>> cf.TimeDuration(2, 'calendar_years') < cf.TimeDuration(
    ...     25, 'calendar_months')
    True
    >>> cf.TimeDuration(2, 'hours') <= cf.TimeDuration(1, 'days')
    True
    >>> cf.TimeDuration(2, 'hours') == cf.TimeDuration(1/12.0, 'days')
    True
    >>> cf.TimeDuration(2, 'days') == cf.TimeDuration(48, 'hours')
    True

    >>> cf.TimeDuration(2, 'hours') <= 2
    True
    >>> 30.5 != cf.TimeDuration(2, 'days')
    True

    >>> cf.TimeDuration(2, 'calendar_years') > numpy.array(1.5)
    True
    >>> type(cf.TimeDuration(2, 'calendar_years') > numpy.array(1.5))
    numpy.bool_
    >>> cf.TimeDuration(2, 'calendar_years') > numpy.array([1.5])
    array([ True])
    >>> numpy.array([[1, 12]]) > cf.TimeDuration(2, 'calendar_months')
    array([[False,  True]])

    >>> cf.TimeDuration(2, 'days') == cf.Data(2)
    <CF Data(): True>
    >>> cf.TimeDuration(2, 'days') == cf.Data([2.], 'days')
    <CF Data(1): [True]>
    >>> cf.Data([[60]], 'seconds') < cf.TimeDuration(2, 'days')
    <CF Data(1, 1): [[True]]>
    >>> cf.Data([1, 12], 'calendar_months') < cf.TimeDuration(
    ...     6, 'calendar_months')
    <CF Data(2): [True, False]>

    **Arithmetic operations**

    Arithmetic operations are defined for `cf.TimeDuration` objects,
    date-time-like objects (such as `cf.Datetime`,
    `datetime.datetime`, etc.), `cf.Data` objects, `numpy` arrays and
    numbers:

    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(28, 'days')
    <CF TimeDuration: P92D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(12, 'hours')
    <CF TimeDuration: P65.0D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'days') + cf.TimeDuration(24, 'hours')
    <CF TimeDuration: P64.5D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'calendar_years') + cf.TimeDuration(
    ...     21, 'calendar_months')
    <CF TimeDuration: P65.75Y (Y-01-01 00:00:00)>

    >>> cf.TimeDuration(30, 'days') + 2
    <CF TimeDuration: P32D (Y-M-D 00:00:00)>
    >>> 4.5 + cf.TimeDuration(30, 'days')
    <CF TimeDuration: P34.5D (Y-M-D 00:00:00)>
    >>> cf.TimeDuration(64, 'calendar_years') - 2.5
    <CF TimeDuration: P61.5Y (Y-01-01 00:00:00)>

    >>> cf.TimeDuration(36, 'hours') / numpy.array(8)
    <CF TimeDuration: 4.5 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'hours') / numpy.array(8.0)
    <CF TimeDuration: 4.5 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'hours') // numpy.array(8.0)
    <CF TimeDuration: 4.0 hours (from Y-M-D h:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') * cf.Data([[2.25]])
    <CF TimeDuration: 81.0 calendar_months (from Y-M-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') // cf.Data([0.825])
    <CF TimeDuration: P43.0M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % 10
    <CF TimeDuration: P6M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % cf.Data(1, 'calendar_year')
    <CF TimeDuration: P0.0M (Y-01-01 00:00:00)>
    >>> cf.TimeDuration(36, 'calendar_months') % cf.Data(2, 'calendar_year')
    <CF TimeDuration: P12.0M (Y-01-01 00:00:00)>

    The in place operators (``+=``, ``//=``, etc.) are supported in a
    similar manner.

    **Attributes**

    ===========  =========================================================
    Attribute    Description
    ===========  =========================================================
    `!duration`  The length of the time duration in a `cf.Data` object
                 with units.
    `!year`      The default year for time interval creation.
    `!month`     The default month for time interval creation.
    `!day`       The default day for time interval creation.
    `!hour`      The default hour for time interval creation.
    `!minute`    The default minute for time interval creation.
    `!second`    The default second for time interval creation.
    ===========  =========================================================


    **Constructors**

    For convenience, the following functions may also be used to
    create time duration objects:

    ========  ============================================================
    Function  Description
    ========  ============================================================
    `cf.Y`    Create a time duration of calendar years.
    `cf.M`    Create a time duration of calendar months.
    `cf.D`    Create a time duration of days.
    `cf.h`    Create a time duration of hours.
    `cf.m`    Create a time duration of minutes.
    `cf.s`    Create a time duration of seconds.
    ========  ============================================================

    .. versionadded:: 1.0

    .. seealso:: `cf.dt`, `cf.Data`, `cf.Datetime`

    '''
    def __init__(self, duration, units=None, month=1, day=1, hour=0,
                 minute=0, second=0):
        '''**Initialization**

    :Parameters:

        duration: data-like
            The length of the time duration.

            A data-like object is any object containing array-like or
            scalar data which could be used to create a `cf.Data`
            object.

            *Parameter example:*
              Instances, ``x``, of following types are all examples of
              data-like objects (because ``cf.Data(x)`` creates a
              valid `cf.Data` object), `int`, `float`, `str`, `tuple`,
              `list`, `numpy.ndarray`, `cf.Data`, `cf.Coordinate`,
              `cf.Field`.

        units: `str` or `cf.Units`, optional
            The units of the time duration. Required if, and only if,
            *duration* is not a `cf.Data` object which already
            contains the units. Units must be one of calendar years,
            calendar months, days, hours, minutes or seconds.

            *Parameter example:*
              ``units='calendar_months'``

            *Parameter example:*
              ``units='days'``

            *Parameter example:*
              ``units=cf.Units('calendar_years')``

         month, day, hour, minute, second: `int` or `None`, optional
            The offset used when creating, with the `bounds` method, a
            time interval containing a given date-time. Only the
            offset elements for units smaller that of the time
            duration are used.

            *Parameter example:*
              >>> cf.TimeDuration(1, 'calendar_month').bounds(
              ...     cf.dt('2000-1-8'))
              (cftime.DatetimeGregorian(2000-01-01 00:00:00),
               cftime.DatetimeGregorian(2000-02-01 00:00:00))
              >>> cf.TimeDuration(1, 'calendar_month', day=15).bounds(
              ...     cf.dt('2000-1-8'))
              (cftime.DatetimeGregorian(1999-12-15 00:00:00),
               cftime.DatetimeGregorian(2000-01-15 00:00:00))
              >>> cf.TimeDuration(
              ...     1, 'calendar_month', month=4, day=30).bounds(
              ...         cf.dt('2000-1-8'))
              (cftime.DatetimeGregorian(1999-12-30 00:00:00),
               cftime.DatetimeGregorian(2000-01-30 00:00:00))

    **Examples:**

    >>> t = cf.TimeDuration(cf.Data(3 , 'calendar_years'))
    >>> t = cf.TimeDuration(cf.Data(12 , 'hours'))
    >>> t = cf.TimeDuration(18 , 'calendar_months')
    >>> t = cf.TimeDuration(30 , 'days')
    >>> t = cf.TimeDuration(1 , 'day', hour=6)

        '''
        if units is not None:
            units = Units(units)
            self.duration = Data(abs(duration))
            if self.duration.Units.istime:
                self.duration.Units = units
            else:
                self.duration = Data(self.duration.array, units)
        else:
            self.duration = abs(Data.asdata(duration))
            units = self.duration.Units
            if not units.istime:
                raise ValueError("Bad units: {!r}".format(units))
        # --- End: if

        if not (units.iscalendartime or units.istime):
            raise ValueError(
                "Can't create {0} of {1}".format(
                    self.__class__.__name__, self.duration)
            )

        duration = self.duration

        offset = [None, month, day, hour, minute, second, 0]
        if units.equivalent(_calendar_years):
            if duration < _one_year:
                offset[1] = None
        else:
            offset[1] = None
            offset[2] = None
            if units <= _hours and duration < _one_day:
                offset[3] = None
                if units <= _minutes and duration < _one_hour:
                    offset[4] = None
                    if units <= _seconds and duration < _one_minute:
                        offset[5] = None
        # --- End: if
        self.offset = Offset(*offset)

# TODO should offset be None for all "higher" units

        self._compound = False

        self._NotImplemented_RHS_Data_op = True

    def __abs__(self):
        '''x.__abs__() <==> abs(x)

    .. versionadded:: 1.4

        '''
        out = self.copy()
        out.duration = abs(self.duration)
        return out

    def __array__(self, *dtype):
        '''TODO
        '''
        return self.duration.__array__(*dtype)

    def __data__(self):
        '''Returns a new reference to the `!duration` attribute.

        '''
        return self.duration

    def __deepcopy__(self, memo):
        '''Used if copy.deepcopy is called

        '''
        return self.copy()

    def __neg__(self):
        '''x.__neg__() <==> -x

    .. versionadded:: 1.4

        '''
        out = self.copy()
        out.duration *= -1
        return out

    def __bool__(self):
        '''Truth value testing and the built-in operation `bool`

    x.__bool__() <==> bool(x)

        '''
        return bool(self.duration)

    def __int__(self):
        '''x.__int__() <==> int(x)

        '''
        return int(self.duration)

    def __repr__(self):
        '''x.__repr__() <==> repr(x)

        '''
        return '<CF {0}: {1}>'.format(self.__class__.__name__, str(self))

    def __str__(self):
        '''x.__str__() <==> str(x)

        '''
        yyy = [x if y is None else '{0:0>2}'.format(y)
               for x, y in zip(('Y', 'M', 'D', 'h', 'm', 's'),
                               self.offset)]

        return '{0} ({1}-{2}-{3} {4}:{5}:{6})'.format(self.iso, *yyy)

    def __ge__(self, other):
        '''The rich comparison operator ``>=``

    x.__ge__(y) <==> x>=y

        '''

        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__ge__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__ge__')

        return NotImplemented

    def __gt__(self, other):
        '''The rich comparison operator ``>``

    x.__gt__(y) <==> x>y

        '''

        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__gt__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__gt__')

        return NotImplemented

    def __le__(self, other):
        '''The rich comparison operator ``<=``

    x.__le__(y) <==> x<=y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__le__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__le__')

        return NotImplemented

    def __lt__(self, other):
        '''The rich comparison operator ``<``

    x.__lt__(y) <==> x<y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__lt__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__lt__')

        return NotImplemented

    def __eq__(self, other):
        '''The rich comparison operator ``==``

    x.__eq__(y) <==> x==y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__eq__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__eq__')

        return NotImplemented

    def __ne__(self, other):
        '''The rich comparison operator ``!=``

    x.__ne__(y) <==> x!=y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return bool(self._binary_operation(other, '__ne__'))

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__ne__')

        return NotImplemented

    def __add__(self, other):
        '''The binary arithmetic operation ``+``

    x.__add__(y) <==> x + y

    .. versionadded:: 1.4

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__add__')

        if hasattr(other, 'timetuple'):
            # other is a date-time object
            try:
                return self._datetime_arithmetic(other, __add__)
            except TypeError:
                return NotImplemented

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__add__')

        return NotImplemented

    def __sub__(self, other):
        '''The binary arithmetic operation ``-``

    x.__sub__(y) <==> x - y

    .. versionadded:: 1.4

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__sub__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__sub__')

        return NotImplemented

    def __mul__(self, other):
        '''The binary arithmetic operation ``*``

    x.__mul__(y) <==> x*y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__mul__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__mul__')

        return NotImplemented

    def __div__(self, other):
        '''The binary arithmetic operation ``/``

    x.__div__(y) <==> x/y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__div__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__div__')

        return NotImplemented

    def __floordiv__(self, other):
        '''The binary arithmetic operation ``//``

    x.__floordiv__(y) <==> x//y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__floordiv__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__floordiv__')

        return NotImplemented

    def __truediv__(self, other):
        '''The binary arithmetic operation ``/`` (true division)

    x.__truediv__(y) <==> x/y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__truediv__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__truediv__')

        return NotImplemented

    def __iadd__(self, other):
        '''The augmented arithmetic assignment ``+=``

    x.__iadd__(y) <==> x+=y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__iadd__', True)

        return NotImplemented

    def __idiv__(self, other):
        '''The augmented arithmetic assignment ``/=``

    x.__idiv__(y) <==> x /= y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__idiv__', True)

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__idiv__', True)

        return NotImplemented

    def __itruediv__(self, other):
        '''The augmented arithmetic assignment ``/=`` (true division)

    x.__truediv__(y) <==> x/y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__itruediv__', True)

        return NotImplemented

    def __ifloordiv__(self, other):
        '''The augmented arithmetic assignment ``//=``

    x.__ifloordiv__(y) <==> x//=y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__ifloordiv__', True)

        return NotImplemented

    def __imul__(self, other):
        '''The augmented arithmetic assignment ``*=``

    x.__imul__(y) <==> x *= y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__imul__', True)

        return NotImplemented

    def __isub__(self, other):
        '''The augmented arithmetic assignment ``-=``

    x.__isub__(y) <==> x -= y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__isub__', True)

        return NotImplemented

    def __imod__(self, other):
        '''The augmented arithmetic assignment ``%=``

    x.__imod__(y) <==> x%=y

        '''
        if isinstance(other, (int, float)):
            return self._binary_operation(other, '__imod__', True)

        if isinstance(other, Data):
            return self._data_binary_operation(other, '__imod__', True)

        return NotImplemented

    def __radd__(self, other):
        '''The binary arithmetic operation ``+`` with reflected operands

    x.__radd__(y) <==> y+x

        '''
        return self + other

    def __rmul__(self, other):
        '''The binary arithmetic operation ``*`` with reflected operands

    x.__rmul__(y) <==> y*x

        '''
        return self * other

    def __rsub__(self, other):
        '''The binary arithmetic operation ``-`` with reflected operands

    x.__rsub__(y) <==> y-x

    .. versionadded:: 1.4

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__rsub__')

        if hasattr(other, 'timetuple'):
            # other is a date-time object
            try:
                return self._datetime_arithmetic(other, __sub__)
            except:
                return NotImplemented

        return NotImplemented

    def __mod__(self, other):
        '''The binary arithmetic operation ``%``

    x.__mod__(y) <==> x % y

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__mod__')

        if isinstance(other, Data):
            return self._data_arithmetic(other, '__mod__')

        return NotImplemented

    def __rmod__(self, other):
        '''The binary arithmetic operation ``%`` with reflected operands

    x.__rmod__(y) <==> y % x

        '''
        if isinstance(other, (self.__class__, int, float)):
            return self._binary_operation(other, '__rmod__')

        return NotImplemented

    # ----------------------------------------------------------------
    # Private methods
    # ----------------------------------------------------------------
    def _binary_operation(self, other, method, inplace=False):
        '''TODO

        '''
        if inplace:
            new = self
        else:
            new = self.copy()

        self_units = self.duration.Units

        other = getattr(other, 'duration', other)

        duration = getattr(new.duration, method)(other)

        duration.squeeze(inplace=True)

        if duration.size != 1:
            raise ValueError(
                "Can't create {} with more than one value: {!r}".format(
                    self.__class__.__name__, duration))

        if duration < 0:
            raise ValueError(
                "Can't create {} with with negative duration {!r}".format(
                    self.__class__.__name__, duration))

        if (not duration.Units.equals(self.duration.Units) and
                method not in _relational_methods):
            # Operator is not one of ==, !=, >=, >, <=, <
            raise ValueError("Can't create {} of {!r}".format(
                self.__class__.__name__, duration.Units))

        if method not in _relational_methods:
            # Operator is not one of ==, !=, >=, >, <=, <
            duration.Units = self_units

        new.duration = duration

        return new

    def _data_binary_operation(self, other, method, inplace=False):
        '''TODO

        '''
        if inplace:
            new = self
        else:
            new = self.copy()

        return getattr(self.duration, method)(other)

    def _datetime_arithmetic(self, other, op):
        '''TODO

    .. versionadded:: 1.4

        '''
        def _dHMS(duration, other, calendar, op):
            units = Units(
                '{0} since {1}'.format(duration.Units.units, other), calendar)
            d = op(Data(0.0, units), duration)
            return d.datetime_array.item(())
        # --- End: def

        duration = self.duration
        units = duration.Units

        if units == _calendar_years:
            months0 = duration.datum() * 12
            months = int(months0)
            if months != months0:
                raise ValueError(
                    "Fractional months not supported for date calculations: "
                    "{}".format(months0)
                )
        elif units == _calendar_months:
            months0 = duration.datum()
            months = int(months0)
            if months != months0:
                raise ValueError(
                    "Fractional months not supported for date calculations: "
                    "{}".format(months0)
                )
        else:
            months = None

        calendar = getattr(other, 'calendar', None)
        if calendar == '':
            calendar = None

        if months is not None:
            y, m = divmod(op(other.month, months), 12)
            if not m:
                y -= 1
                m = 12

            y = other.year + y

            max_days = self.days_in_month(y, m, calendar)
            d = other.day
            if d > max_days:
                d = max_days

            return other.replace(year=y, month=m, day=d)
        else:
            return _dHMS(duration, other, calendar, op)

    def _data_arithmetic(self, other, method, inplace=False):
        '''TODO

        '''
        try:
            dt = other.datetime_array
        except ValueError:
            return self._binary_operation(other, method, inplace=inplace)
        else:
            out = []
            for d in dt.flat:
                if d is numpy.ma.masked:
                    out.append(None)
                else:
                    out.append(getattr(self, method)(d))
            # --- End: for

            dt[...] = numpy.reshape(out, dt.shape)

            return Data(dt, units=other.Units)

    def _offset(self, dt):
        '''TODO

    .. versionadded:: 1.4

    :Parameters:

        dt: `cf.Datetime` TODO

    :Returns:

        `Datetime` TODO

        '''
        return cf_dt(*[(i if j is None else j)
                       for i, j in zip(elements(dt), self.offset)],
                     calendar=getattr(dt, 'calendar', None))

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def iso(self):
        '''Return the time duration as an ISO 8601-like time duration string.

    .. versionadded:: 1.0

    **Examples:**

    >>> cf.TimeDuration(45, 'days').iso
    'P45D'
    >>> cf.TimeDuration(10, 'calendar_years').iso
    'P10Y'
    >>> cf.s(5.67).iso
    'PT5.67S'
    >>> cf.M(18.5).iso
    'P18.5M'

        '''
        duration = self.duration
        units = duration.Units

        if units.equals(_calendar_months):
            return 'P{0}M'.format(duration.datum())
        if units.equals(_calendar_years):
            return 'P{0}Y'.format(duration.datum())
        if units.equals(_days):
            return 'P{0}D'.format(duration.datum())
        if units.equals(_hours):
            return 'PT{0}H'.format(duration.datum())
        if units.equals(_minutes):
            return 'PT{0}M'.format(duration.datum())
        if units.equals(_seconds):
            return 'PT{0}S'.format(duration.datum())

        raise ValueError(
            "Bad {0} units: {1!r}".format(self.__class__.__name__, units))

    @property
    def isint(self):
        '''True if the time duration is a whole number.

    .. versionadded:: 1.0

    **Examples:**

    >>> cf.TimeDuration(2, 'hours').isint
    True
    >>> cf.TimeDuration(2.0, 'hours').isint
    True
    >>> cf.TimeDuration(2.5, 'hours').isint
    False

        '''
        duration = self.duration

        if duration.dtype.kind == 'i':
            return True

        duration = duration.datum()
        return int(duration) == float(duration)

    @property
    def Units(self):
        '''The units of the time duration.

    .. versionadded:: 1.0

    **Examples:**

    >>> cf.TimeDuration(3, 'days').Units
    <CF Units: days>

    >>> t = cf.TimeDuration(cf.Data(12, 'calendar_months'))
    >>> t.Units
    <CF Units: calendar_months>
    >>> t.Units = cf.Units('calendar_years')
    >>> t.Units
    <CF Units: calendar_years>
    >>> t
    <CF TimeDuration: 1.0 calendar_years (from Y-01-01 00:00:00)>

        '''
        return self.duration.Units

    @Units.setter
    def Units(self, value):
        duration = getattr(self, 'duration', None)
        if duration is None:
            raise AttributeError(
                "Can't set units when there is no duration attribute")

        self.duration.Units = Units(value)

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def copy(self):
        '''Return a deep copy.

    ``t.copy()`` is equivalent to ``copy.deepcopy(t)``.

    .. versionadded:: 1.0

    :Returns:

            The deep copy.

    **Examples:**

    >>> u = t.copy()

        '''
        new = TimeDuration.__new__(TimeDuration)
        new.__dict__ = self.__dict__.copy()
        new.duration = self.duration.copy()
        return new

    @classmethod
    def days_in_month(cls, year, month, calendar=None, leap_month=2,
                      month_lengths=None):
        '''The number of days in a specific month in a specific year in a
    specific calendar.

    .. versionadded:: 1.4

    :Parameters:

        year: `int`

        month: `int`

        calendar: `str`, optional
            By default, calendar is the mixed Gregorian/Julian
            calendar as defined by Udunits.

        leap_month: `int`, optional
            By default, the leap month is 2.

        month_lengths: sequence of `int`, optional
            By default, *month_lengths* is ``[31, 28, 31, 30, 31, 30,
            31, 31, 30, 31, 30, 31]``.

    :Returns:

        `int`
            The number of days in the specified month.

    **Examples:**

    >>> cf.TimeDuration.days_in_month(2004, 2, calendar='360_day')
    30

        '''
        if month_lengths is None:
            month_lengths = _default_month_lengths
        elif len(month_lengths) != 12:
            raise ValueError(
                "month_lengths must be a sequence of 12 elements")

        month1 = month - 1

        if calendar in [
                None, 'standard', 'gregorian', 'proleptic_gregorian', '']:
            length = month_lengths[month1]
            if (month == leap_month and
                    (year % 400 == 0 or (year % 100 != 0 and year % 4 == 0))):
                length += 1
        elif calendar == '360_day':
            length = 30
        elif calendar in ('noleap', '365_day'):
            length = month_lengths[month1]
        elif calendar in ('all_leap', '366_day'):
            length = month_lengths[month1]
            if month == leap_month:
                length += 1
        elif calendar == 'julian':
            length = month_lengths[month1]
            if month == leap_month and year % 4 == 0:
                length += 1

        return length

    @_deprecated_kwarg_check('traceback')
    def equals(self, other, rtol=None, atol=None, verbose=False,
               traceback=False):
        '''True if two time durations are equal.

    .. versionadded:: 1.0

    .. seealso:: `equivalent`

    :Parameters:

        other:
            The object to compare for equality.

        atol: `float`, optional
            The absolute tolerance for all numerical comparisons, By
            default the value returned by the `ATOL` function is used.

        rtol: `float`, optional
            The relative tolerance for all numerical comparisons, By
            default the value returned by the `RTOL` function is used.

        traceback: `bool`, optional
            If True then print a traceback highlighting where the two
            instances differ.

    :Returns:

        `bool`
            Whether or not the two instances are equal.

    **Examples:**

    >>> t = cf.TimeDuration(36, 'calendar_months')
    >>> u = cf.TimeDuration(3, 'calendar_years')
    >>> t == u
    True
    >>> t.equals(u, traceback=True)
    TimeDuration: Different durations: <CF Data: 36 calendar_months>, <CF Data: 3 calendar_years>
    False

        '''
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if verbose:  # pragma: no cover
                print(
                    "%s: Different type: %s" % (
                        self.__class__.__name__, other.__class__.__name__)
                )  # pragma: no cover
            return False

        self__dict__ = self.__dict__.copy()
        other__dict__ = other.__dict__.copy()

        d0 = self__dict__.pop('duration', None)
        d1 = other__dict__.pop('duration', None)

        if not d0.equals(d1):
            if verbose:  # pragma: no cover
                print(
                    "%s: Different durations: %r, %r" % (
                        self.__class__.__name__, d0, d1)
                )  # pragma: no cover
            return False

        if self__dict__ != other__dict__:
            if verbose:  # pragma: no cover
                print(
                    "%s: Different The default date-time elements: "
                    "%r != %r" % (
                        self.__class__.__name__, self__dict__, other__dict__)
                )  # pragma: no cover
            return False

        return True

    @_deprecated_kwarg_check('traceback')
    def equivalent(self, other, rtol=None, atol=None, verbose=True,
                   traceback=False):
        '''True if two time durations are logically equivalent.

    .. versionadded:: 1.0

    .. seealso:: `equals`

    :Parameters:

        other:
            The object to compare for equivalence.

        atol: `float`, optional
            The absolute tolerance for all numerical comparisons, By
            default the value returned by the `ATOL` function is used.

        rtol: `float`, optional
            The relative tolerance for all numerical comparisons, By
            default the value returned by the `RTOL` function is used.

        traceback: `bool`, optional
            If True then print a traceback highlighting where the two
            instances differ.

    :Returns:

        `bool`
            Whether or not the two instances logically equivalent.

    **Examples:**

    >>> t = cf.TimeDuration(36, 'calendar_months')
    >>> u = cf.TimeDuration(3, 'calendar_years')
    >>> t == u
    True
    >>> t.equivalent(u)
    True
    >>> t.equals(u, traceback=True)
    TimeDuration: Different durations: <CF Data: 12 calendar_months>, <CF Data: 1 calendar_years>
    False

        '''
        # Check each instance's id
        if self is other:
            return True

        # Check that each instance is the same type
        if self.__class__ != other.__class__:
            if verbose:  # pragma: no cover
                print(
                    "%s: Different type: %s" % (
                        self.__class__.__name__, other.__class__.__name__)
                )  # pragma: no cover
            return False

        self__dict__ = self.__dict__.copy()
        other__dict__ = other.__dict__.copy()

        d0 = self__dict__.pop('duration', None)
        d1 = other__dict__.pop('duration', None)
        if d0 != d0:
            if verbose:  # pragma: no cover
                print(
                    "%s: Non-equivalent durations: %r, %r" % (
                        self.__class__.__name__, d0, d1)
                )  # pragma: no cover
            return False

        if self__dict__ != other__dict__:
            if verbose:  # pragma: no cover
                print(
                    "%s: Non-equivalent default date-time elements: "
                    "%r != %r" % (
                        self.__class__.__name__, self__dict__, other__dict__)
                )  # pragma: no cover
            return False

        return True

    def inspect(self):
        '''Inspect the attributes.

    .. versionadded:: 1.0

    .. seealso:: `cf.inspect`

    :Returns:

        `None`

        '''
        print(cf_inspect(self))  # pragma: no cover

    def interval(self, dt, end=False, iso=None):  # calendar=None,
        '''Return a time interval of exactly the time duration.

    The start (or end, if the *end* parameter is True) date-time of
    the time interval is determined by the date-time given by the *dt*
    parameter.

    .. versionadded:: 1.0

    .. seealso:: `bounds`

    :Parameters:

        dt:
            The date-time. One of:

            * A `str` specifying an ISO 8601-like date-time string (in
              which non-Gregorian calendar dates are allowed).

            * `datetime.datetime or sublcass `cftime.datetime` (such
              as `Datetime360Day`).

        end: `bool`, optional
            If True then the date-time given by the *year*, *month*,
            *day*, *hour*, *minute* and *second* parameters defines
            the end of the time interval. By default it defines the
            start of the time interval.

        iso: `str`, optional
            Return the time interval as an ISO 8601 time interval
            string rather than the default of a tuple of date-time
            objects. Valid values are (with example outputs for the
            time interval "3 years from 2007-03-01 13:00:00"):

              ========================  =============================================
              iso                       Example output
              ========================  =============================================
              ``'start and end'``       ``'2007-03-01 13:00:00/2010-03-01 13:00:00'``
              ``'start and duration'``  ``'2007-03-01 13:00:00/P3Y'``
              ``'duration and end'``    ``'P3Y/2010-03-01 13:00:00'``
              ========================  =============================================

    :Returns:

            The date-times at each end of the time interval. The first
            date-time is always earlier than or equal to the second
            date-time. If *iso* has been set then an ISO 8601 time
            interval string is returned instead of a tuple.

    **Examples:**

    >>> t = cf.TimeDuration(6, 'calendar_months')
    >>> t
    <CF TimeDuration: P6M (Y-M-01 00:00:00)>
    >>> t.interval(cf.dt(1999, 12))
    (cftime.DatetimeGregorian(1999-12-01 00:00:00),
     cftime.DatetimeGregorian(2000-06-01 00:00:00))

    >>> t = cf.TimeDuration(5, 'days', hour=6)
    >>> t
    <CF TimeDuration: P5D (Y-M-D 06:00:00)>
    >>> t.interval(cf.dt(2004, 3, 2), end=True)
    (cftime.DatetimeGregorian(2004-02-26 00:00:00),
     cftime.DatetimeGregorian(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='noleap'), end=True)
    (cftime.DatetimeNoLeap(2004-02-25 00:00:00),
     cftime.DatetimeNoLeap(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'), end=True)
    (cftime.Datetime360Day(2004-02-27 00:00:00),
     cftime.Datetime360Day(2004-03-02 00:00:00))
    >>> t.interval(cf.dt(2004, 3, 2, calendar='360_day'),
    ...            end=True, iso='start and duration')
    '2004-02-27 00:00:00/P5D'

    Create `cf.Query` objects for a time interval - one including both
    bounds and one which excludes the upper bound:

    >>> t = cf.TimeDuration(2, 'calendar_years')
    >>> interval = t.interval(cf.dt(1999, 12))
    >>> c = cf.wi(*interval)
    >>> c
    <CF Query: (wi [cftime.DatetimeGregorian(1999-12-01 00:00:00), cftime.DatetimeGregorian(2001-12-01 00:00:00)])>
    >>> c == cf.dt('2001-1-1', calendar='gregorian')
    True

    Create a `cf.Query` object which may be used to test where a time
    coordinate object's bounds lie inside a time interval:

    >>> t = cf.TimeDuration(1, 'calendar_months')
    >>> c = cf.cellwi(*t.interval(cf.dt(2000, 1), end=True))
    >>> c
    <CF Query: [lower_bounds(ge 1999-12-01 00:00:00) & upper_bounds(le 2000-01-01 00:00:00)]>

    Create ISO 8601 time interval strings:

    >>> t = cf.TimeDuration(6, 'calendar_years')
    >>> t.interval(cf.dt(1999, 12), end=True, iso='start and end')
    '1993-12-01 00:00:00/1999-12-01 00:00:00'
    >>> t.interval(cf.dt(1999, 12), end=True, iso='start and duration')
    '1993-12-01 00:00:00/P6Y'
    >>> t.interval(cf.dt(1999, 12), end=True, iso='duration and end')
    'P6Y/1999-12-01 00:00:00'

        '''
        def _dHMS(duration, dt, end):
            calendar = dt.calendar
            if not calendar:
                calendar = None

            units = Units('{0} since {1}'.format(duration.Units.units, dt),
                          calendar)
            dt1 = Data(0.0, units)

            if not end:
                dt1 += duration
            else:
                dt1 -= duration

            dt1 = dt1.datetime_array.item(())

            if not end:
                return dt, dt1  # dt.copy(), dt1
            else:
                return dt1, dt  # dt1, dt.copy()
        # --- End: def

        calendar = getattr(dt, 'calendar', _default_calendar)
        if calendar == '':
            calendar = _default_calendar

        dt = cf_dt(dt, calendar=calendar)

        duration = self.duration
        units = duration.Units

        if units == _calendar_years:
            months = duration.datum() * 12
            int_months = int(months)
            if int_months != months:
                raise ValueError(
                    "Can't create a time interval of a non-integer number "
                    "of calendar months: {0}".format(months)
                )
        elif units == _calendar_months:
            months = duration.datum()
            int_months = int(months)
            if int_months != months:
                raise ValueError(
                    "Can't create a time interval of a non-integer number "
                    "of calendar months: {0}".format(months)
                )
        else:
            int_months = None

        if int_months is not None:
            if not end:
                y, month1 = divmod(dt.month + int_months, 12)

                if not month1:
                    y -= 1
                    month1 = 12

                year1 = dt.year + y

                max_days = self.days_in_month(year1, month1, calendar)

                day1 = dt.day
                if day1 > max_days:
                    day1 = max_days

                dt0 = dt  # .copy()
                dt1 = dt.replace(year=year1, month=month1, day=day1)
            else:
                y, month0 = divmod(dt.month - int_months, 12)

                if not month0:
                    y -= 1
                    month0 = 12

                year0 = dt.year + y

                max_days = self.days_in_month(year0, month0, calendar)
                day0 = dt.day
                if day0 > max_days:
                    day0 = max_days

                dt0 = dt.replace(year=year0, month=month0, day=day0)
                dt1 = dt  # 0.copy()
        else:
            dt0, dt1 = _dHMS(duration, dt, end)

        if not iso:
            return dt0, dt1

        if iso == 'start and end':
            return '{0}/{1}'.format(dt0, dt1)
        if iso == 'start and duration':
            return '{0}/{1}'.format(dt0, self.iso)
        if iso == 'duration and end':
            return '{0}/{1}'.format(self.iso, dt1)

    def bounds(self, dt, direction=True):
        '''Return a time interval containing a date-time.

    The interval spans the time duration and starts and ends at
    date-times consistent with the time duration's offset.

    .. versionadded:: 1.2.3

    .. seealso:: `cf.dt`, `cf.Datetime`, `interval` TODO

    :Parameters:

        dt: date-time-like
            The date-time to be contained by the interval. *dt* may be
            any date-time-like object, such as `cf.Datetime`,
            `datetime.datetime`, `netCDF4.netcdftime.datetime`, etc.

            *Parameter example:*
              To find bounds around 1999-16-1 in the Gregorian
              calendar you could use ``dt=cf.dt(1999, 1, 16)`` or
              ``dt=datetime.datetime(1999, 1, 16)`` (See `cf.dt` for
              details).

        direction: `bool`, optional
            If `False` then the bounds are decreasing. By default the
            bounds are increasing. Note that ``t.bounds(dt,
            direction=False)`` is equivalent to
            ``t.bounds(dt)[::-1]``.

    :Returns:

        `tuple`
            The two bounds.

    **Examples:**

    TODO

    >>> t = cf.M()
    >>> t.bounds(cf.dt(2000, 1, 1))
    (cftime.DatetimeGregorian(2000-01-01 00:00:00),
     cftime.DatetimeGregorian(2000-02-01 00:00:00))

    >>> t = cf.M(1)
    >>> t.bounds(cf.dt(2000, 3, 1))
    (cftime.DatetimeGregorian(2000-03-01 00:00:00),
     cftime.DatetimeGregorian(2000-04-01 00:00:00))

    >>> t = cf.M(1, day=15)
    >>> t.bounds(cf.dt(2000, 3, 1))
    (cftime.DatetimeGregorian(2000-02-15 00:00:00),
     cftime.DatetimeGregorian(2000-03-15 00:00:00))

    >>> t = cf.M(2, day=15)
    >>> t.bounds(cf.dt(2000, 3, 1), direction=False)
    (cftime.DatetimeGregorian(2000-03-15 00:00:00),
     cftime.DatetimeGregorian(2000-01-15 00:00:00))

        '''
        abs_self = abs(self)

        calendar = getattr(dt, 'calendar', _default_calendar)
        if calendar == '':
            calendar = _default_calendar

        dt = cf_dt(dt, calendar=calendar)

        dt0 = self._offset(dt)

        if dt0 > dt:
            dt1 = dt0
            dt0 = dt1 - abs_self
        else:
            dt1 = dt0 + abs_self

        if direction:
            return (dt0, dt1)
        else:
            return (dt1, dt0)

    def is_day_factor(self):
        '''Return True if an integer multiple of the time duration is equal
    to one day.

    .. versionadded:: 1.0

    :Returns:

        `bool`

    **Examples:**

    >>> cf.TimeDuration(0.5, 'days').is_day_factor()
    True

    >>> cf.TimeDuration(1, 'days').is_day_factor()
    True
    >>> cf.TimeDuration(0.25, 'days').is_day_factor()
    True
    >>> cf.TimeDuration(0.3, 'days').is_day_factor()
    False
    >>> cf.TimeDuration(2, 'days').is_day_factor()
    False

    >>> cf.TimeDuration(24, 'hours').is_day_factor()
    True
    >>> cf.TimeDuration(6, 'hours').is_day_factor()
    True
    >>> cf.TimeDuration(7, 'hours').is_day_factor()
    False
    >>> cf.TimeDuration(27, 'hours').is_day_factor()
    False

    >>> cf.TimeDuration(1440, 'minutes').is_day_factor()
    True
    >>> cf.TimeDuration(15, 'minutes').is_day_factor()
    True
    >>> cf.TimeDuration(17, 'minutes').is_day_factor()
    False
    >>> cf.TimeDuration(2007, 'minutes').is_day_factor()
    False

    >>> cf.TimeDuration(86400, 'seconds').is_day_factor()
    True
    >>> cf.TimeDuration(45, 'seconds').is_day_factor()
    True
    >>> cf.TimeDuration(47, 'seconds').is_day_factor()
    False
    >>> cf.TimeDuration(86401, 'seconds').is_day_factor()
    False

    >>> cf.TimeDuration(1, 'calendar_months').is_day_factor()
    False
    >>> cf.TimeDuration(1, 'calendar_years').is_day_factor()
    False

        '''
        try:
            return not Data(1, 'day') % self.duration
        except ValueError:
            return False


# --- End: class


def Y(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of calendar years in a `cf.TimeDuration`
    object.

    ``cf.Y()`` is equivalent to ``cf.TimeDuration(1,
    'calendar_year')``.

    ``cf.Y(duration, *args, **kwargs)`` is equivalent to
    ``cf.TimeDuration(duration, 'calendar_year', *args, **kwargs)``.

    .. versionadded:: 1.0

    .. seealso:: `cf.M`, `cf.D`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of calendar years in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.Y(month=12)`` is equivalent to ``cf.TimeDuration(1,
              'calendar_years', month=12)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.Y()
    <CF TimeDuration: P1Y (Y-01-01 00:00:00)>
    >>> cf.Y(10, month=12)
    <CF TimeDuration: P10Y (Y-12-01 00:00:00)>
    >>> cf.Y(15, month=4, day=2, hour=12, minute=30, second=2)
    <CF TimeDuration: P15Y (Y-04-02 12:30:02)>
    >>> cf.Y(0)
    <CF TimeDuration: P0Y (Y-M-01 00:00:00)>

    '''
    return TimeDuration(duration, 'calendar_years', month=month,
                        day=day, hour=hour, minute=minute,
                        second=second)


def M(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of calendar months in a `cf.TimeDuration`
    object.

    ``cf.M()`` is equivalent to ``cf.TimeDuration(1, 'calendar_month')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.D`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of calendar months in the time duration.

        year, month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.M(day=16)`` is equivalent to ``cf.TimeDuration(1,
              'calendar_months', day=16)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.M()
    <CF TimeDuration: P1M (Y-M-01 00:00:00)>
    >>> cf.M(3, day=16)
    <CF TimeDuration: P3M (Y-M-16 00:00:00)>
    >>> cf.M(24, day=2, hour=12, minute=30, second=2)
    <CF TimeDuration: P24M (Y-01-02 12:30:02)>
    >>> cf.M(0)
    <CF TimeDuration: P0M (Y-M-01 00:00:00)>

    '''
    return TimeDuration(duration, 'calendar_months', month=month, day=day,
                        hour=hour, minute=minute, second=second)


def D(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of days in a `cf.TimeDuration` object.

    ``cf.D()`` is equivalent to ``cf.TimeDuration(1, 'day')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.h`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of days in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and end
            of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for details.

            *Parameter example:*
               ``cf.D(hour=12)`` is equivalent to ``cf.TimeDuration(1,
               'day', hour=12)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.D()
    <CF TimeDuration: P1D (Y-M-D 00:00:00)>
    >>> cf.D(5, hour=12)
    <CF TimeDuration: P5D (Y-M-D 12:00:00)>
    >>> cf.D(48.5, minute=30)
    <CF TimeDuration: P48.5D (Y-M-D 00:30:00)>
    >>> cf.D(0.25, hour=6, minute=30, second=20)
    <CF TimeDuration: P0.25D (Y-M-D 06:30:20)>
    >>> cf.D(0)
    <CF TimeDuration: P0D (Y-M-D 00:00:00)>

    '''
    return TimeDuration(duration, 'days', month=month, day=day,
                        hour=hour, minute=minute, second=second)


def h(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of hours in a `cf.TimeDuration` object.

    ``cf.h()`` is equivalent to ``cf.TimeDuration(1, 'hour')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.m`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        year, month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.h(minute=30)`` is equivalent to
              ``cf.TimeDuration(1, 'hour', minute=30)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.h()
    <CF TimeDuration: PT1H (Y-M-D h:00:00)>
    >>> cf.h(3, minute=15)
    <CF TimeDuration: PT3H (Y-M-D h:15:00)>
    >>> cf.h(0.5)
    <CF TimeDuration: PT0.5H (Y-M-D h:00:00)>
    >>> cf.h(6.5, minute=15, second=45)
    <CF TimeDuration: PT6.5H (Y-M-D h:15:45)>
    >>> cf.h(0)
    <CF TimeDuration: PT0H (Y-M-D h:00:00)>

    '''
    return TimeDuration(duration, 'hours', month=month, day=day,
                        hour=hour, minute=minute, second=second)


def m(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of minutes in a `cf.TimeDuration` object.

    ``cf.m()`` is equivalent to ``cf.TimeDuration(1, 'minute')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.h`, `cf.s`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining when a time
            interval based on this time duration begins or ends. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.m(second=30)`` is equivalent to
              ``cf.TimeDuration(1, 'minute', second=30)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.m()
    <CF TimeDuration: PT1M (Y-M-D h:m:00)>
    >>> cf.m(30, second=15)
    <CF TimeDuration: PT30M (Y-M-D h:m:15)>
    >>> cf.m(0.5)
    <CF TimeDuration: PT0.5M (Y-M-D h:m:00)>
    >>> cf.m(2.5, second=45)
    <CF TimeDuration: PT2.5M (Y-M-D h:m:45)>
    >>> cf.m(0)
    <CF TimeDuration: PT0M (Y-M-D h:m:00)>

    '''
    return TimeDuration(duration, 'minutes', month=month, day=day,
                        hour=hour, minute=minute, second=second)


def s(duration=1, month=1, day=1, hour=0, minute=0, second=0):
    '''Return a time duration of seconds in a `cf.TimeDuration` object.

    ``cf.s()`` is equivalent to ``cf.TimeDuration(1, 'second')``.

    .. versionadded:: 1.0

    .. seealso:: `cf.Y`, `cf.M`, `cf.D`, `cf.h`, `cf.m`

    :Parameters:

        duration: number, optional
            The number of hours in the time duration.

        month, day, hour, minute, second: `int`, optional
            The default date-time elements for defining the start and
            end of a time interval based on this time duration. See
            `cf.TimeDuration` and `cf.TimeDuration.interval` for
            details.

            *Parameter example:*
              ``cf.s(hour=6)`` is equivalent to ``cf.TimeDuration(1,
              'seconds', hour=6)``.

    :Returns:

        `TimeDuration`
            The new `cf.TimeDuration` object.

    **Examples:**

    >>> cf.s()
    <CF TimeDuration: PT1S (Y-M-D h:m:s)>
    >>> cf.s().interval(cf.dt(1999, 12, 1))
    (cftime.DatetimeGregorian(1999-12-01 00:00:00),
     cftime.DatetimeGregorian(1999-12-01 00:00:01))
    >>> cf.s(30)
    <CF TimeDuration: PT30S (Y-M-D h:m:s)>
    >>> cf.s(0.5)
    <CF TimeDuration: PT0.5S (Y-M-D h:m:s)>
    >>> cf.s(12.25)
    <CF TimeDuration: PT12.25S (Y-M-D h:m:s)>
    >>> cf.s(2.5, year=1999, hour=12)
    cf.s(30, month=2, hour=12)
    <CF TimeDuration: PT30S (Y-M-D h:m:s)>
    >>> cf.s(0)
    <CF TimeDuration: PT0S (Y-M-D h:m:s)>

    '''
    return TimeDuration(duration, 'seconds', month=month, day=day,
                        hour=hour, minute=minute, second=second)



from copy import copy

from .functions import (_DEPRECATION_ERROR,
                        _DEPRECATION_ERROR_KWARGS,
                        _DEPRECATION_ERROR_METHOD,
                        _DEPRECATION_ERROR_DICT)

from .decorators import _deprecated_kwarg_check


class FieldList(list):
    '''An ordered sequence of fields.

    Each element of a field list is a field construct.

    A field list supports the python list-like operations (such as
    indexing and methods like `!append`).

    >>> fl = cf.FieldList()
    >>> len(fl)
    0
    >>> f
    <CF Field: air_temperaturetime(12), latitude(73), longitude(96) K>
    >>> fl = cf.FieldList(f)
    >>> len(fl)
    1
    >>> fl = cf.FieldList([f, f])
    >>> len(fl)
    2
    >>> fl = cf.FieldList(cf.FieldList([f] * 3))
    >>> len(fl)
    3
    >>> len(fl + fl)
    6

    These methods provide functionality similar to that of a
    :ref:`built-in list <python:tut-morelists>`. The main difference
    is that when a field element needs to be assesed for equality its
    `~cf.Field.equals` method is used, rather than the ``==``
    operator.

    '''
    def __init__(self, fields=None):
        '''**Initialization**

    :Parameters:

        fields: (sequence of) `Field`, optional
             Create a new field list with these fields.

        '''
        if fields is not None:
            if getattr(fields, 'construct_type', None) == 'field':
                self.append(fields)
            else:
                self.extend(fields)

    def __call__(self, *identities):
        '''Alias for `cf.FieldList.select_by_identity`.

        '''
        return self.select_by_identity(*identities)

    def __repr__(self):
        '''Called by the `repr` built-in function.

    x.__repr__() <==> repr(x)

        '''
        out = [repr(f) for f in self]
        out = ',\n '.join(out)
        return '['+out+']'

    # ----------------------------------------------------------------
    # Overloaded list methods
    # ----------------------------------------------------------------
    def __add__(self, x):
        '''The binary arithmetic operation ``+``

    f.__add__(x) <==> f + x

    :Returns:

        `FieldList`

    **Examples:**

    >>> h = f + g
    >>> f += g

        '''
        return type(self)(list.__add__(self, x))

    def __contains__(self, y):
        '''Called to implement membership test operators.

    x.__contains__(y) <==> y in x

    Each field in the field list is compared with the field's
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    Note that ``y in fl`` is equivalent to ``any(f.equals(y) for f in fl)``.

        '''
        for f in self:
            if f.equals(y):
                return True
        # --- End: for

        return False

    def __mul__(self, x):
        '''The binary arithmetic operation ``*``

    f.__mul__(x) <==> f * x

    :Returns:

        `FieldList`

    **Examples:**

    >>> h = f * 2
    >>> f *= 2

        '''
        return type(self)(list.__mul__(self, x))

    def __eq__(self, other):
        '''The rich comparison operator ``==``

    f.__eq__(x) <==> f==x

    Each field in the field list is compared with the field's
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    Note that ``f==x`` is equivalent to ``f.equals(x)``.

    :Returns:

        `bool`

        '''
        return self.equals(other)

    def __getslice__(self, i, j):
        '''Called to implement evaluation of f[i:j]

    f.__getslice__(i, j) <==> f[i:j]

    :Returns:

        `FieldList`

    **Examples:**

    >>> g = f[0:1]
    >>> g = f[1:-4]
    >>> g = f[:1]
    >>> g = f[1:]

        '''
        return type(self)(list.__getslice__(self, i, j))

    def __getitem__(self, index):
        '''Called to implement evaluation of f[index]

    f.__getitem_(index) <==> f[index]

    :Returns:

        `Field` or `FieldList`
            If *index* is an integer then a field construct is
            returned. If *index* is a slice then a field list is returned,
            which may be empty.

    **Examples:**

    >>> g = f[0]
    >>> g = f[-1:-4:-1]
    >>> g = f[2:2:2]

        '''
        out = list.__getitem__(self, index)

        if isinstance(out, list):
            return type(self)(out)

        return out

    def __ne__(self, other):
        '''The rich comparison operator ``!=``

    f.__ne__(x) <==> f!=x

    Each field in the field list is compared with the field's
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    Note that ``f!=x`` is equivalent to ``not f.equals(x)``.

    :Returns:

        `bool`

        '''
        return not self.equals(other)

    # ???
    __len__     = list.__len__
    __setitem__ = list.__setitem__
    append      = list.append
    extend      = list.extend
    insert      = list.insert
    pop         = list.pop
    reverse     = list.reverse
    sort        = list.sort

    # ----------------------------------------------------------------
    # Methods
    # ----------------------------------------------------------------
    def close(self):
        '''Close all files referenced by each field construct.

    Note that a closed file will be automatically reopened if its
    contents are subsequently required.

    :Returns:

        `None`

    **Examples:**

    >>> fl.close()

        '''
        for f in self:
            f.close()

    def count(self, value):
        '''Return number of occurrences of value

    Each field in the field list is compared with the field's
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    Note that ``fl.count(value)`` is equivalent to
    ``sum(f.equals(value) for f in fl)``.

    .. seealso:: `cf.Field.equals`, `list.count`

    **Examples:**

    >>> f = cf.FieldList([a, b, c, a])
    >>> f.count(a)
    2
    >>> f.count(b)
    1
    >>> f.count(a+1)
    0

        '''
        return len([None for f in self if f.equals(value)])

    def index(self, value, start=0, stop=None):
        '''Return first index of value.

    Each field in the field list is compared with the field's
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    An exception is raised if there is no such field.

    .. seealso:: `list.index`

        '''
        if start < 0:
            start = len(self) + start

        if stop is None:
            stop = len(self)
        elif stop < 0:
            stop = len(self) + stop

        for i, f in enumerate(self[start:stop]):
            if f.equals(value):
                return i + start
        # --- End: for

        raise ValueError(
            "{0!r} is not in {1}".format(value, self.__class__.__name__))

    def remove(self, value):
        '''Remove first occurrence of value.

    Each field in the field list is compared with its
    `~cf.Field.equals` method, as opposed to the ``==`` operator.

    .. seealso:: `list.remove`

        '''
        for i, f in enumerate(self):
            if f.equals(value):
                del self[i]
                return
        # --- End: for

        raise ValueError(
            "{0}.remove(x): x not in {0}".format(self.__class__.__name__))

    def sort(self, key=None, reverse=False):
        '''Sort of the field list in place.

    By default the field list is sorted by the identities of its field
    construct elements.

    The sort is stable.

    .. versionadded:: 1.0.4

    .. seealso:: `reverse`

    :Parameters:

        key: function, optional
            Specify a function of one argument that is used to extract
            a comparison key from each field construct. By default the
            field list is sorted by field identity, i.e. the default
            value of *key* is ``lambda f: f.identity()``.

        reverse: `bool`, optional
            If set to `True`, then the field list elements are sorted
            as if each comparison were reversed.

    :Returns:

        `None`

    **Examples:**

    >>> fl
    [<CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>,
     <CF Field: ocean_meridional_overturning_streamfunction(time(12), region(4), depth(40), latitude(180)) m3 s-1>,
     <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>,
     <CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>]
    >>> fl.sort()
    >>> fl
    [<CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>,
     <CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>,
     <CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>,
     <CF Field: ocean_meridional_overturning_streamfunction(time(12), region(4), depth(40), latitude(180)) m3 s-1>]
    >>> fl.sort(reverse=True)
    >>> fl
    [<CF Field: ocean_meridional_overturning_streamfunction(time(12), region(4), depth(40), latitude(180)) m3 s-1>,
     <CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>,
     <CF Field: eastward_wind(time(3), air_pressure(5), grid_latitude(110), grid_longitude(106)) m s-1>,
     <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>]

    >>> [f.datum(0) for f in fl]
    [masked,
     -0.12850454449653625,
     -0.12850454449653625,
     236.51275634765625]
    >>> fl.sort(key=lambda f: f.datum(0), reverse=True)
    >>> [f.datum(0) for f in fl]
    [masked,
     236.51275634765625,
     -0.12850454449653625,
     -0.12850454449653625]

    >>> from operator import attrgetter
    >>> [f.long_name for f in fl]
    ['Meridional Overturning Streamfunction',
     'U COMPNT OF WIND ON PRESSURE LEVELS',
     'U COMPNT OF WIND ON PRESSURE LEVELS',
     'air_temperature']
    >>> fl.sort(key=attrgetter('long_name'))
    >>> [f.long_name for f in fl]
    ['air_temperature',
     'Meridional Overturning Streamfunction',
     'U COMPNT OF WIND ON PRESSURE LEVELS',
     'U COMPNT OF WIND ON PRESSURE LEVELS']

        '''
        if key is None:
            key = lambda f: f.identity()

        return super().sort(key=key, reverse=reverse)

    def __deepcopy__(self, memo):
        '''Called by the `copy.deepcopy` standard library function.

        '''
        return self.copy()

    def concatenate(self, axis=0, _preserve=True):
        '''Join the sequence of fields together.

    This is different to `cf.aggregate` because it does not account
    for all metadata. For example, it assumes that the axis order is
    the same in each field.

    .. versionadded:: 1.0

    .. seealso:: `cf.aggregate`, `Data.concatenate`

    :Parameters:

        axis: `int`, optional
            TODO

    :Returns:

        `Field`
            TODO

        '''
        return self[0].concatenate(self, axis=axis, _preserve=_preserve)

    def copy(self, data=True):
        '''Return a deep copy.

    ``f.copy()`` is equivalent to ``copy.deepcopy(f)``.

    :Returns:

            The deep copy.

    **Examples:**

    >>> g = f.copy()
    >>> g is f
    False
    >>> f.equals(g)
    True
    >>> import copy
    >>> h = copy.deepcopy(f)
    >>> h is f
    False
    >>> f.equals(g)
    True

        '''
        return type(self)([f.copy(data=data) for f in self])

    @_deprecated_kwarg_check('traceback')
    def equals(self, other, rtol=None, atol=None, verbose=False,
               ignore_data_type=False, ignore_fill_value=False,
               ignore_properties=(), ignore_compression=False,
               ignore_type=False, ignore=(), traceback=False,
               unordered=False):
        '''Whether two field lists are the same.

    Equality requires the two field lists to have the same length and
    for the field construct elements to be equal pair-wise, using
    their `~cf.Field.equals` methods.

    Any type of object may be tested but, in general, equality is only
    possible with another field list, or a subclass of one. See the
    *ignore_type* parameter.

    Equality is between teo field constructs is strict by
    default. This means that for two field constructs to be considered
    equal they must have corresponding metadata constructs and for
    each pair of constructs:

    * the same descriptive properties must be present, with the same
      values and data types, and vector-valued properties must also
      have same the size and be element-wise equal (see the
      *ignore_properties* and *ignore_data_type* parameters), and

    ..

    * if there are data arrays then they must have same shape and data
      type, the same missing data mask, and be element-wise equal (see
      the *ignore_data_type* parameter).

    Two real numbers ``x`` and ``y`` are considered equal if
    ``|x-y|<=atol+rtol|y|``, where ``atol`` (the tolerance on absolute
    differences) and ``rtol`` (the tolerance on relative differences)
    are positive, typically very small numbers. See the *atol* and
    *rtol* parameters.

    If data arrays are compressed then the compression type and the
    underlying compressed arrays must be the same, as well as the
    arrays in their uncompressed forms. See the *ignore_compression*
    parameter.

    NetCDF elements, such as netCDF variable and dimension names, do
    not constitute part of the CF data model and so are not checked on
    any construct.

    :Parameters:
        other:
            The object to compare for equality.

        atol: float, optional
            The tolerance on absolute differences between real
            numbers. The default value is set by the `cfdm.ATOL`
            function.

        rtol: float, optional
            The tolerance on relative differences between real
            numbers. The default value is set by the `cfdm.RTOL`
            function.

        ignore_fill_value: `bool`, optional
            If `True` then the "_FillValue" and "missing_value"
            properties are omitted from the comparison, for the field
            construct and metadata constructs.

        verbose: `bool`, optional
            If `True` then print information about differences that lead
            to inequality.

        ignore_properties: sequence of `str`, optional
            The names of properties of the field construct (not the
            metadata constructs) to omit from the comparison. Note
            that the "Conventions" property is always omitted by
            default.

        ignore_data_type: `bool`, optional
            If `True` then ignore the data types in all numerical
            comparisons. By default different numerical data types
            imply inequality, regardless of whether the elements are
            within the tolerance for equality.

        ignore_compression: `bool`, optional
            If `True` then any compression applied to underlying arrays
            is ignored and only uncompressed arrays are tested for
            equality. By default the compression type and, if
            applicable, the underlying compressed arrays must be the
            same, as well as the arrays in their uncompressed forms

        ignore_type: `bool`, optional
            Any type of object may be tested but, in general, equality
            is only possible with another field list, or a subclass of
            one. If *ignore_type* is True then
            ``FieldList(source=other)`` is tested, rather than the
            ``other`` defined by the *other* parameter.

        unordered: `bool`, optional
            TODO

    :Returns:

        `bool`
            Whether the two field lists are equal.

    **Examples:**

    >>> fl.equals(fl)
    True
    >>> fl.equals(fl.copy())
    True
    >>> fl.equals(fl[:])
    True
    >>> fl.equals('not a FieldList instance')
    False

        '''
        if ignore:
            _DEPRECATION_ERROR_KWARGS(
                self, 'equals', {'ignore': ignore},
                "Use keyword 'ignore_properties' instead."
            )  # pragma: no cover

        # Check for object identity
        if self is other:
            return True

        # Check that each object is of compatible type
        if ignore_type:
            if not isinstance(other, self.__class__):
                other = type(self)(source=other, copy=False)
        elif not isinstance(other, self.__class__):
            if verbose:
                print(
                    "{0}: Incompatible type: {1}".format(
                        self.__class__.__name__, other.__class__.__name__)
                )  # pragma: no cover
            return False

        # Check that there are equal numbers of fields
        len_self = len(self)
        if len_self != len(other):
            if verbose:
                print(
                    "{0}: Different numbers of field construct: "
                    "{1}, {2}".format(
                        self.__class__.__name__,
                        len_self, len(other))
                )  # pragma: no cover
            return False

        if not unordered or len_self == 1:
            # ----------------------------------------------------
            # Check the lists pair-wise
            # ----------------------------------------------------
            for i, (f, g) in enumerate(zip(self, other)):
                if not f.equals(g, rtol=rtol, atol=atol,
                                ignore_fill_value=ignore_fill_value,
                                ignore_properties=ignore_properties,
                                ignore_compression=ignore_compression,
                                ignore_data_type=ignore_data_type,
                                ignore_type=ignore_type,
                                verbose=verbose):
                    if verbose:
                        print(
                            "{0}: Different field constructs at element {1}: "
                            "{2!r}, {3!r}".format(
                                self.__class__.__name__, i, f, g)
                        )  # pragma: no cover
                    return False
        else:
            # ----------------------------------------------------
            # Check the lists set-wise
            # ----------------------------------------------------
            # Group the variables by identity
            self_identity = {}
            for f in self:
                self_identity.setdefault(f.identity(), []).append(f)

            other_identity = {}
            for f in other:
                other_identity.setdefault(f.identity(), []).append(f)

            # Check that there are the same identities
            if set(self_identity) != set(other_identity):
                if verbose:
                    print("{}: Different sets of identities: {}, {}".format(
                        self.__class__.__name__,
                        set(self_identity),
                        set(other_identity)))  # pragma: no cover
                return False

            # Check that there are the same number of variables
            # for each identity
            for identity, fl in self_identity.items():
                gl = other_identity[identity]
                if len(fl) != len(gl):
                    if verbose:
                        print(
                            "{0}: Different numbers of {1!r} {2}s: "
                            "{3}, {4}".format(
                                self.__class__.__name__,
                                identity,
                                fl[0].__class__.__name__,
                                len(fl), len(gl)
                            )
                        )  # pragma: no cover
                    return False
            # --- End: for

            # For each identity, check that there are matching pairs
            # of equal fields.
            for identity, fl in self_identity.items():
                gl = other_identity[identity]

                for f in fl:
                    found_match = False
                    for i, g in enumerate(gl):
                        if f.equals(g, rtol=rtol, atol=atol,
                                    ignore_fill_value=ignore_fill_value,
                                    ignore_properties=ignore_properties,
                                    ignore_compression=ignore_compression,
                                    ignore_data_type=ignore_data_type,
                                    ignore_type=ignore_type,
                                    verbose=verbose):
                            found_match = True
                            del gl[i]
                            break
                # --- End: for

                if not found_match:
                    if verbose:
                        print(
                            "{0}: No {1} equal to: {2!r}".format(
                                self.__class__.__name__,
                                g.__class__.__name__, f)
                        )  # pragma: no cover
                    return False
        # --- End: if

        # ------------------------------------------------------------
        # Still here? Then the field lists are equal
        # ------------------------------------------------------------
        return True

    def select_by_construct(self, *identities, OR=False, **conditions):
        '''Select field constructs by metadata constructs.

    To find the inverse of the selection, use a list comprehension
    with the `~cf.Field.match_by_construct` method of the field
    constucts. For example, to select all field constructs that do
    *not* have a "latitude" metadata construct:

       >>> gl = cf.FieldList(f for f in fl
       ...                   if not f.match_by_constructs('latitude'))

    .. note:: The API changed at version 3.1.0

    .. versionadded:: 3.0.0

    .. seealso: `select`, `__call__`, `select_by_units`,
                `select_by_naxes`, `select_by_rank`,
                `select_by_property`, `cf.Field.match_by_identity`,
                `cf.Field.subspace`

    :Parameters:

        identities: optional
            Identify the metadata constructs that have any of the
            given identities or construct keys.

            A construct identity is specified by a string
            (e.g. ``'latitude'``, ``'long_name=time'``,
            ``'ncvar%lat'``, etc.); or a compiled regular expression
            (e.g. ``re.compile('^atmosphere')``) that selects the
            relevant constructs whose identities match via
            `re.search`.

            Each construct has a number of identities, and is selected
            if any of them match any of those provided. A construct's
            identities are those returned by its `!identities`
            method. In the following example, the construct ``x`` has
            six identities:

               >>> x.identities()
               ['time',
                'long_name=Time',
                'foo=bar',
                'standard_name=time',
                'ncvar%t',
                'T']

            A construct key may optionally have the ``'key%'``
            prefix. For example ``'dimensioncoordinate2'`` and
            ``'key%dimensioncoordinate2'`` are both acceptable keys.

            Note that in the output of a `print` call or `!dump`
            method, a construct is always described by one of its
            identities, and so this description may always be used as
            an *identity* argument.

            If a cell method construct identity is given (such as
            ``'method:mean'``) then it will only be compared with the
            most recently applied cell method operation.

            Alternatively, one or more cell method constucts may be
            identified in a single string with a CF-netCDF cell
            methods-like syntax for describing both the collapse
            dimensions, the collapse method, and any cell method
            construct qualifiers. If N cell methods are described in
            this way then they will collectively identify the N most
            recently applied cell method operations. For example,
            ``'T: maximum within years T: mean over years'`` will be
            compared with the most two most recently applied cell
            method operations.

            *Parameter example:*
              ``'measure:area'``

            *Parameter example:*
              ``'latitude'``

            *Parameter example:*
              ``'long_name=Longitude'``

            *Parameter example:*
              ``'domainancillary2', 'ncvar%areacello'``

        conditions: optional
            Identify the metadata constructs that have any of the
            given identities or construct keys, and whose data satisfy
            conditions.

            A construct identity or construct key (as defined by the
            *identities* parameter) is given as a keyword name and a
            condition on its data is given as the keyword value.

            The condition is satisfied if any of its data values
            equals the value provided.

            *Parameter example:*
              ``longitude=180.0``

            *Parameter example:*
              ``time=cf.dt('1959-12-16')``

            *Parameter example:*
              ``latitude=cf.ge(0)``

            *Parameter example:*
              ``latitude=cf.ge(0), air_pressure=500``

            *Parameter example:*
              ``**{'latitude': cf.ge(0), 'long_name=soil_level': 4}``

        OR: `bool`, optional
            If True then return `True` if at least one metadata
            construct matches at least one of the criteria given by
            the *identities* or *conditions* arguments. By default
            `True` is only returned if the field constructs matches
            each of the given criteria.

        mode: deprecated at version 3.1.0
            Use the *OR* parameter instead.

        constructs: deprecated at version 3.1.0

    :Returns:

        `bool`
            The matching field constructs.

    **Examples:**

        TODO

        '''
#        if constructs:
#            for key, value in constructs.items():
#                if value is None:
#                    message = ("Since its value is None, use {!r} as a "
#                               "positional argument instead".format(value))
#                else:
#                    message = ("Evaluating criteria on data values is no "
#                               "longer possible with this method.")
#
#                _DEPRECATION_ERROR_KWARGS(self, 'select_by_construct',
#                                          kwargs={key: value},
#                                          message=message,
#                                          version='3.1.0') # pragma: no cover
#        # --- End: if

        if identities:
            if identities[0] == 'or':
                _DEPRECATION_ERROR_ARG(
                    self, 'select_by_construct', 'or',
                    message="Use 'OR=True' instead.", version='3.1.0'
                )  # pragma: no cover

            if identities[0] == 'and':
                _DEPRECATION_ERROR_ARG(
                    self, 'select_by_construct', 'and',
                    message="Use 'OR=False' instead.", version='3.1.0'
                )  # pragma: no cover
        # --- End: if

        return type(self)(
            f for f in self
            if f.match_by_construct(*identities, OR=OR, **conditions)
        )

    def select_by_identity(self, *identities):
        '''Select field constructs by identity.

    To find the inverse of the selection, use a list comprehension
    with the `~cf.Field.match_by_identity` method of the field
    constucts. For example, to select all field constructs whose
    identity is *not* ``'air_temperature'``:

       >>> gl = cf.FieldList(f for f in fl
       ...                   if not f.match_by_identity('air_temperature'))

    .. versionadded:: 3.0.0

    .. seealso:: `select`, `__call__`, `select_by_units`,
                 `select_by_construct`, `select_by_naxes`,
                 `select_by_rank`, `select_by_property`,
                 `cf.Field.match_by_identity`

    :Parameters:

        identities: optional
            Select field constructs. By default all field constructs
            are selected. May be one or more of:

              * The identity of a field construct.

            A construct identity is specified by a string (e.g.
            ``'air_temperature'``, ``'long_name=Air Temperature',
            ``'ncvar%tas'``, etc.); or a compiled regular expression
            (e.g. ``re.compile('^air_')``) that selects the relevant
            constructs whose identities match via `re.search`.

            Each construct has a number of identities, and is selected
            if any of them match any of those provided. A construct's
            identities are those returned by its `!identities`
            method. In the following example, the construct ``x`` has
            five identities:

               >>> x.identities()
               ['air_temperature',
                'long_name=Air Temperature',
                'foo=bar',
                'standard_name=air_temperature',
                'ncvar%tas']

            Note that in the output of a `print` call or `!dump`
            method, a construct is always described by one of its
            identities, and so this description may always be used as
            an *identities* argument.

    :Returns:

        `FieldList`
            The matching field constructs.

    **Examples:**

    >>> fl
    [<CF Field: specific_humidity(latitude(73), longitude(96)) 1>,
     <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>]
    >>> fl.select('air_temperature')
    [<CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>]

        '''
        return type(self)(f for f in self if f.match_by_identity(*identities))

    def select_by_naxes(self, *naxes):
        '''Select field constructs by property.

    To find the inverse of the selection, use a list comprehension
    with `~cf.Field.match_by_naxes` method of the field constucts. For
    example, to select all field constructs which do *not* have
    3-dimensional data:

       >>> gl = cf.FieldList(f for f in fl if not f.match_by_naxes(3))

    .. versionadded:: 3.0.0

    .. seealso:: `select`, `select_by_identity`,
                 `select_by_construct`, `select_by_property`,
                 `select_by_rank`, `select_by_units`

    :Parameters:

        naxes: optional
            Select field constructs whose data spans a particular
            number of domain axis constructs.

            A number of domain axis constructs is given by an `int`.

            If no numbers are provided then all field constructs are
            selected.

    :Returns:

        `FieldList`
            The matching field constructs.

    **Examples:**

    TODO

        '''
        return type(self)(f for f in self if f.match_by_naxes(*naxes))

    def select_by_rank(self, *ranks):
        '''Select field constructs by the number of domain axis constructs.

    .. versionadded:: 3.0.0

    .. seealso: `select`, `__call__`, `select_by_units`,
                `select_by_naxes`, `select_by_construct`,
                `select_by_property`, `cf.Field.match_by_identity`

    :Parameters:

        ranks: optional
            Define conditions on the number of domain axis constructs.

            A condition is one of:

              * `int`
              * a `Query` object

            The condition is satisfied if the number of domain axis
            constructs equals the condition value.

            *Parameter example:*
              To see if the field construct has 4 domain axis
              constructs: ``4``

            *Parameter example:*
              To see if the field construct has at least 3 domain axis
              constructs: ``cf.ge(3)``

    :Returns:

        `bool`
            The matching field constructs.

    **Examples:**

        TODO
        '''

        return type(self)(f for f in self if f.match_by_rank(*ranks))

    def select_by_ncvar(self, *rank):
        '''Select field constructs by netCDF variable name.

    To find the inverse of the selection, use a list comprehension
    with `~cf.Field.match_by_ncvar` method of the field constucts. For
    example, to select all field constructs which do *not* have a
    netCDF name of 'tas':

       >>> gl = cf.FieldList(f for f in fl if not f.match_by_ncvar('tas'))

    .. versionadded:: 3.0.0

    .. seealso:: `select`, `select_by_identity`,
                 `select_by_construct`, `select_by_naxes`,
                 `select_by_rank`, `select_by_units`

    :Parameters:

        ncvars: optional
            Select field constructs. May be one or more:

              * The netCDF name of a field construct.

            A field construct is selected if it matches any of the
            given names.

            A netCDF variable name is specified by a string (e.g.
            ``'tas'``, etc.); a `Query` object
            (e.g. ``cf.eq('tas')``); or a compiled regular expression
            (e.g. ``re.compile('^air_')``) that selects the field
            constructs whose netCDF variable names match via
            `re.search`.

            If no netCDF variable names are provided then all field
            are selected.

    :Returns:

        `FieldList`
            The matching field constructs.

    **Examples:**

    TODO

        '''
        return type(self)(f for f in self if f.match_by_ncvar(*ncvars))

    def select_by_property(self, *mode, **properties):
        '''Select field constructs by property.

    To find the inverse of the selection, use a list comprehension
    with `~cf.Field.match_by_property` method of the field
    constucts. For example, to select all field constructs which do
    *not* have a long_name property of 'Air Pressure':

       >>> gl = cf.FieldList(f for f in fl if not
       ...                   f.match_by_property(long_name='Air Pressure))

    .. versionadded:: 3.0.0

    .. seealso:: `select`, `select_by_identity`,
                 `select_by_construct`, `select_by_naxes`,
                 `select_by_rank`, `select_by_units`

    :Parameters:

        mode: optional
            Define the behaviour when multiple properties are
            provided.

            By default (or if the *mode* parameter is ``'and'``) a
            field construct is selected if it matches all of the given
            properties, but if the *mode* parameter is ``'or'`` then a
            field construct will be selected when at least one of its
            properties matches.

        properties: optional
            Select field constructs. May be one or more of:

              * The property of a field construct.

            By default a field construct is selected if it matches all
            of the given properties, but it may alternatively be
            selected when at least one of its properties matches (see
            the *mode* positional parameter).

            A property value is given by a keyword parameter of the
            property name. The value may be a scalar or vector
            (e.g. ``'air_temperature'``, ``4``, ``['foo', 'bar']``);
            or a compiled regular expression
            (e.g. ``re.compile('^ocean')``), for which all constructs
            whose methods match (via `re.search`) are selected.

    :Returns:

        `FieldList`
            The matching field constructs.

    **Examples:**

    TODO

        '''
        return type(self)(
            f for f in self if f.match_by_property(*mode, **properties))

    def select_by_units(self, *units, exact=True):
        '''Select field constructs by units.

    To find the inverse of the selection, use a list comprehension
    with `~cf.Field.match_by_units` method of the field constucts. For
    example, to select all field constructs whose units are *not*
    ``'km'``:

       >>> gl = cf.FieldList(f for f in fl if not f.match_by_units('km'))

    .. versionadded:: 3.0.0

    .. seealso:: `select`, `select_by_identity`,
                 `select_by_construct`, `select_by_naxes`,
                 `select_by_rank`, `select_by_property`

    :Parameters:

        units: optional
            Select field constructs. By default all field constructs
            are selected. May be one or more of:

              * The units of a field construct.

            Units are specified by a string or compiled regular
            expression (e.g. 'km', 'm s-1', ``re.compile('^kilo')``,
            etc.) or a `Units` object (e.g. ``Units('km')``,
            ``Units('m s-1')``, etc.).

        exact: `bool`, optional
            If `False` then select field constructs whose units are
            equivalent to any of those given by *units*. For example,
            metres and are equivelent to kilometres. By default, field
            constructs whose units are exactly one of those given by
            *units* are selected. Note that the format of the units is
            not important, i.e. 'm' is exactly the same as 'metres'
            for this purpose.

    :Returns:

        `FieldList`
            The matching field constructs.

    **Examples:**

    >>> gl = fl.select_by_units('metres')
    >>> gl = fl.select_by_units('m')
    >>> gl = fl.select_by_units('m', 'kilogram')
    >>> gl = fl.select_by_units(Units('m'))
    >>> gl = fl.select_by_units('km', exact=False)
    >>> gl = fl.select_by_units(Units('km'), exact=False)
    >>> gl = fl.select_by_units(re.compile('^met'))
    >>> gl = fl.select_by_units(Units('km'))
    >>> gl = fl.select_by_units(Units('kg m-2'))

        '''
        return type(self)(f for f in self
                          if f.match_by_units(*units, exact=exact))

    def select_field(self, identity, default=ValueError()):
        '''Select a unique field construct by its identity.

    .. versionadded:: 3.0.4

    .. seealso:: `select`, `select_by_identity`

    :Parameters:

        identity:
            Select the field construct. May be:

              * The identity of a field construct.

            A construct identity is specified by a string (e.g.
            ``'air_temperature'``, ``'long_name=Air Temperature',
            ``'ncvar%tas'``, etc.); or a compiled regular expression
            (e.g. ``re.compile('^air_')``) that selects the relevant
            constructs whose identities match via `re.search`.

            Each construct has a number of identities, and is selected
            if any of them match any of those provided. A construct's
            identities are those returned by its `!identities`
            method. In the following example, the construct ``x`` has
            five identities:

               >>> x.identities()
               ['air_temperature',
                'long_name=Air Temperature',
                'foo=bar',
                'standard_name=air_temperature',
                'ncvar%tas']

            Note that in the output of a `print` call or `!dump`
            method, a construct is always described by one of its
            identities, and so this description may always be used as
            an *identity* argument.

        default: optional
            Return the value of the *default* parameter if a unique
            field construct can not be found. If set to an `Exception`
            instance then it will be raised instead.

    :Returns:

        `Field`
            The unique matching field construct.

    **Examples:**

    >>> fl
    [<CF Field: specific_humidity(latitude(73), longitude(96)) 1>,
     <CF Field: specific_humidity(latitude(73), longitude(96)) 1>,
     <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>]
    >>> fl.select_field('air_temperature')
    <CF Field: air_temperature(time(12), latitude(64), longitude(128)) K>
    >>> f.select_field('specific_humidity')
    ValueError: Multiple fields found
    >>> f.select_field('specific_humidity', 'No unique field')
    'No unique field'
    >>> f.select_field('snowfall_amount')
    ValueError: No fields found

        '''
        out = self.select_by_identity(identity)

        if not out or len(out) > 1:
            if isinstance(default, Exception):
                if not default.args:
                    if not out:
                        message = "No fields found"
                    else:
                        message = "Multiple fields found"

                    default = copy(default)
                    default.args = (message,)

                raise default

            return default

        return out[0]

    # ----------------------------------------------------------------
    # Aliases
    # ----------------------------------------------------------------
    def select(self, *identities, **kwargs):
        '''Alias of `cf.FieldList.select_by_identity`.

    To find the inverse of the selection, use a list comprehension
    with the `~cf.Field.match_by_identity` method of the field
    constucts. For example, to select all field constructs whose
    identity is *not* ``'air_temperature'``:

       >>> gl = cf.FieldList(f for f in fl
       ...                   if not f.match_by_identity('air_temperature'))

    .. seealso:: `select_by_identity`, `select_field`

        '''
        if kwargs:
            _DEPRECATION_ERROR_KWARGS(
                self, 'select', kwargs,
                "Use methods 'select_by_units', 'select_by_construct', "
                "'select_by_properties', 'select_by_naxes', 'select_by_rank' "
                "instead."
            )  # pragma: no cover

        if identities and isinstance(identities[0], (list, tuple, set)):
            _DEPRECATION_ERROR(
                "Use of a {!r} for identities has been deprecated. Use the "
                "* operator to unpack the arguments instead.".format(
                    identities[0].__class__.__name__)
            )  # pragma: no cover

        for i in identities:
            if isinstance(i, dict):
                _DEPRECATION_ERROR_DICT(
                    "Use methods 'select_by_units', 'select_by_construct', "
                    "'select_by_properties', 'select_by_naxes', "
                    "'select_by_rank' instead."
                )  # pragma: no cover

            if isinstance(i, str) and ':' in i:
                error = True
                if '=' in i:
                    index0 = i.index('=')
                    index1 = i.index(':')
                    error = index0 > index1

                if error:
                    _DEPRECATION_ERROR(
                        "The identity format {!r} has been deprecated at "
                        "version 3.0.0. Try {!r} instead.".format(
                            i,  i.replace(':', '=', 1))
                    )  # pragma: no cover
        # --- End: for

        return self.select_by_identity(*identities)

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def _parameters(self, d):
        '''Deprecated at version 3.0.0.

        '''
        _DEPRECATION_ERROR_METHOD(self, '_parameters')  # pragma: no cover

    def _deprecated_method(self, name):
        '''Deprecated at version 3.0.0.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, '_deprecated_method')  # pragma: no cover

    def set_equals(self, other, rtol=None, atol=None,
                   ignore_data_type=False, ignore_fill_value=False,
                   ignore_properties=(), ignore_compression=False,
                   ignore_type=False, traceback=False):
        '''Deprecated at version 3.0.0. Use method 'equals' with
    unordered=True instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'set_equals',
            "Use method 'equals' with unordered=True instead."
        )  # pragma: no cover

    def select1(self, *args, **kwargs):
        '''Deprecated at version 3.0.0. Use method 'fl.select_field' instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'select1', "Use method 'fl.select_field' instead."
        )  # pragma: no cover


# --- End: class



import datetime
import os
import sys
import unittest

import numpy

import cf

class create_fieldTest(unittest.TestCase):
    filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            'test_file.nc')
    chunk_sizes = (17, 34, 300, 100000)[::-1]

    def test_create_field(self):
        # Dimension coordinates
        dim1 = cf.DimensionCoordinate(data=cf.Data(numpy.arange(10.), 'degrees'))
        dim1.standard_name = 'grid_latitude'

        dim0 = cf.DimensionCoordinate(data=cf.Data(numpy.arange(9.) + 20, 'degrees'))
        dim0.standard_name = 'grid_longitude'
        dim0.data[-1] += 5
        bounds = cf.Data(numpy.array([dim0.data.array-0.5, dim0.data.array+0.5]).transpose((1,0)))
        bounds[-2,1] = 30
        bounds[-1,:] = [30, 36]
        dim0.set_bounds(cf.Bounds(data=bounds))

        dim2 = cf.DimensionCoordinate(data=cf.Data([1.5]),
                                      bounds=cf.Bounds(data=cf.Data([[1, 2.]])))
        dim2.standard_name = 'atmosphere_hybrid_height_coordinate'

        # Auxiliary coordinates
        ak = cf.DomainAncillary(data=cf.Data([10.], 'm'))
        ak.id = 'atmosphere_hybrid_height_coordinate_ak'
        bounds = cf.Bounds(data=cf.Data([[5, 15.]], units=ak.Units))
        ak.set_bounds(bounds)

        bk = cf.DomainAncillary(data=cf.Data([20.]))
        bk.id = 'atmosphere_hybrid_height_coordinate_bk'
        bounds = cf.Bounds(data=cf.Data([[14, 26.]]))
        bk.set_bounds(bounds)

        aux2 = cf.AuxiliaryCoordinate(
            data=cf.Data(numpy.arange(-45, 45, dtype='int32').reshape(10, 9),
                         units='degree_N'))
        aux2.standard_name = 'latitude'

        aux3 = cf.AuxiliaryCoordinate(
            data=cf.Data(numpy.arange(60, 150, dtype='int32').reshape(9, 10),
                         units='degreesE'))
        aux3.standard_name = 'longitude'

        aux4 = cf.AuxiliaryCoordinate(
            data=cf.Data(numpy.array(['alpha','beta','gamma','delta','epsilon',
                                      'zeta','eta','theta','iota','kappa'], dtype='S')))
        aux4.standard_name = 'greek_letters'
        aux4[0] = cf.masked


        # Cell measures
        msr0 = cf.CellMeasure(
            data=cf.Data(1+numpy.arange(90.).reshape(9, 10)*1234, 'km 2'))
        msr0.measure = 'area'

        # Data
        data = cf.Data(numpy.arange(90.).reshape(10, 9), 'm s-1')

        properties = {'standard_name': 'eastward_wind'}

        f = cf.Field(properties=properties)

        axisX = f.set_construct(cf.DomainAxis(9))
        axisY = f.set_construct(cf.DomainAxis(10))
        axisZ = f.set_construct(cf.DomainAxis(1))

        f.set_data(data)

        x = f.set_construct(dim0)
        y = f.set_construct(dim1, axes=[axisY])
        z = f.set_construct(dim2, axes=[axisZ])

        lat = f.set_construct(aux2)
        lon = f.set_construct(aux3, axes=['X', axisY])
        f.set_construct(aux4, axes=['Y'])

        ak = f.set_construct(ak, axes=['Z'])
        bk = f.set_construct(bk, axes=[axisZ])

        # Coordinate references
        coordinate_conversion =cf.CoordinateConversion(
            parameters={'grid_mapping_name': 'rotated_latitude_longitude',
                        'grid_north_pole_latitude': 38.0,
                        'grid_north_pole_longitude': 190.0})
        ref0 = cf.CoordinateReference(coordinate_conversion=coordinate_conversion,
                                      coordinates=[x, y, lat, lon])


        f.set_construct(msr0, axes=[axisX, 'Y'])

        f.set_construct(ref0)

        orog = cf.DomainAncillary()
        orog.standard_name = 'surface_altitude'
        orog.set_data(cf.Data(f.array*2, 'm'))
        orog.transpose([1, 0], inplace=True)

        orog_key = f.set_construct(orog, axes=['X', axisY])

        coordinate_conversion = cf.CoordinateConversion(
            parameters={'standard_name': 'atmosphere_hybrid_height_coordinate'},
            domain_ancillaries={'orog': orog_key, 'a': ak, 'b': bk})
        ref1 = cf.CoordinateReference(coordinate_conversion=coordinate_conversion,
                                      coordinates=[z])


        f.set_construct(ref1)

        # Field ancillary variables
        g = cf.FieldAncillary()
        g.set_data(f.data)
        g.transpose([1, 0], inplace=True)
        g.standard_name = 'ancillary0'
        g *= 0.01
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f.data)
        g.standard_name = 'ancillary1'
        g *= 0.01
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f[0, :].data)
        g.squeeze(inplace=True)
        g.standard_name = 'ancillary2'
        g *= 0.001
        f.set_construct(g)

        g = cf.FieldAncillary()
        g.set_data(f[:, 0].data)
        g.squeeze(inplace=True)
        g.standard_name = 'ancillary3'
        g *= 0.001
        f.set_construct(g)


        f.flag_values = [1,2,4]
        f.flag_meanings = ['a', 'bb', 'ccc']

        for cm in cf.CellMethod.create('grid_longitude: mean grid_latitude: max'):
            f.set_construct(cm)


#        print(f.constructs.data_axes())
#        print(repr(f.constructs))
#        print(f.cell_measures)
#        print(f.constructs)
#
#        f.dump()

        # Write the file, and read it in
        cf.write(f, self.filename, verbose=False, string=True)

        g = cf.read(self.filename, squeeze=True, verbose=False)[0]

        self.assertTrue(g.equals(f, verbose=False), "Field not equal to itself read back in")

        x = g.dump(display=False)
        x = f.dump(display=False)


# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)



import numpy

from numpy import array       as numpy_array
from numpy import asscalar    as numpy_asscalar
from numpy import ndenumerate as numpy_ndenumerate
from numpy import empty       as numpy_empty
from numpy import expand_dims as numpy_expand_dims
from numpy import squeeze     as numpy_squeeze

from copy import deepcopy

from operator import mul

from .partition import Partition

from ..functions import (_DEPRECATION_WARNING_METHOD,
                         _DEPRECATION_ERROR_METHOD)

from ..decorators import (_inplace_enabled,
                          _inplace_enabled_define_and_cleanup)


_empty_matrix = numpy_empty((), dtype=object)


class PartitionMatrix:
    '''

A hyperrectangular partition matrix of a master data array.

Each of elements (called partitions) span all or part of exactly one
sub-array of the master data array.

Normal numpy basic and advanced indexing is supported, but size 1
dimensions are always removed from the output array, i.e. a partition
rather than a partition matrix is returned if the output array has
size 1.


**Attributes**

==========  ===========================================================
Attribute   Description
==========  ===========================================================
`!axes`
`!matrix`
`!ndim`     The number of partition dimensions in the partition matrix.
`!shape`    List of the partition matrix's dimension sizes.
`!size`     The number of partitions in the partition matrix.
==========  ===========================================================

'''

    def __init__(self, matrix, axes):
        '''**Initialization**

    :Parameters:

        matrix: `numpy.ndarray`
            An array of Partition objects.

        axes: `list`
            The identities of the partition axes of the partition
            array. If the partition matrix is a scalar array then it
            is an empty list. DO NOT UPDATE INPLACE.

    **Examples:**

    >>> pm = PartitionMatrix(
    ...          numpy.array(Partition(
    ...              location = [(0, 1), (2, 4)],
    ...              shape = [1, 2],
    ...              _dimensions = ['dim2', 'dim0'],
    ...              Units = cf.Units('m'),
    ...              part = [],
    ...              data = numpy.array([[5, 6], [7, 8]])
    ...          ), dtype=object),
    ...          axes=[]
    ...      )

        '''
        self.matrix = matrix
        self.axes = axes

    def __deepcopy__(self, memo):
        '''Used if copy.deepcopy is called on the variable.

        '''
        return self.copy()

    def __getitem__(self, indices):
        '''x.__getitem__(indices) <==> x[indices]

    Normal numpy basic and advanced indexing is supported, but size 1
    dimensions are always removed from the output array, i.e. a
    partition rather than a partition matrix is returned if the output
    array has size 1.

    Returns either a partition or a partition matrix.

    **Examples:**

    >>> pm.shape
    (5, 3)
    >>> pm[0, 1]
    <cf.data.partition.Partition at 0x1934c80>
    >>> pm[:, 1]
    <CF PartitionMatrix: 1 partition dimensions>
    >>> pm[:, 1].shape
    (5,)
    >>> pm[1:4, slice(2, 0, -1)].shape
    (3, 2)

    >>> pm.shape
    ()
    >>> pm[()]
    <cf.data.partition.Partition at 0x1934c80>
    >>> pm[...]
    <cf.data.partition.Partition at 0x1934c80>

        '''
        out = self.matrix[indices]

        if isinstance(out, Partition):
            return out

        if out.size == 1:
            return self.matrix.item()

        axes = [axis for axis, n in zip(self.axes, out.shape) if n != 1]

        return type(self)(numpy_squeeze(out), axes)

    def __repr__(self):
        '''x.__repr__() <==> repr(x)

        '''
        return '<CF %s: %s>' % (self.__class__.__name__, self.shape)

    def __setitem__(self, indices, value):
        '''x.__setitem__(indices, y) <==> x[indices]=y

    Indices must be an integer, a slice object or a tuple. If a slice
    object is given then the value being assigned must be an
    iterable. If a tuple of integers (or slices equivalent to an
    integer) is given then there must be one index per partition
    matrix dimension.

    **Examples:**

    >>> pm.shape
    (3,)
    >>> pm[2] = p1
    >>> pm[:] = [p1, p2, p3]

    >>> pm.shape
    (2, 3)
    >>> pm[0, 2] = p1


    >>> pm.shape
    ()
    >>> pm[()] = p1
    >>> pm[...] = p1

        '''
        self.matrix[indices] = value

    def __str__(self):
        '''x.__str__() <==> str(x)

        '''
        return str(self.matrix)
        out = []
        for partition in self.matrix.flat:
            out.append(str(partition))

        return '\n'.join(out)

    def change_axis_names(self, axis_map):
        '''Change the axis names.

    The axis names are arbitrary, so mapping them to another arbitrary
    collection does not change the data array values, units, nor axis
    order.

    :Parameters:

        axis_map: `dict`

    :Returns:

        `None`

        '''
        # Partition dimensions
        axes = self.axes
        self.axes = [axis_map[axis] for axis in axes]

        # Partitions. Note that a partition may have dimensions which
        # are not in self.axes and that these must also be in
        # axis_name_map.
        for partition in self.matrix.flat:
            partition.change_axis_names(axis_map)

    # ----------------------------------------------------------------
    # Attributes
    # ----------------------------------------------------------------
    @property
    def flat(self):
        '''A flat iterator over the partitions in the partition matrix.

    **Examples:**

    >>> pm.shape
    [2, 2]
    >>> for partition in pm.flat:
    ...     print(repr(partition.Units))
    ...
    <CF Units: m s-1>
    <CF Units: km hr-1>
    <CF Units: miles day-1>
    <CF Units: mm minute-1>

    >>> pm.flat
    <numpy.flatiter at 0x1e14840>

    >>> flat = pm.flat
    >>> next(flat)
    <Partition at 0x1934c80>
    >>> next(flat)
    <Partition at 0x784b347>

        '''
        return self.matrix.flat

    @property
    def ndim(self):
        '''The number of partition dimensions in the partition matrix.

    Not to be confused with the number of dimensions of the master
    data array.

    **Examples:**

    >>> pm.shape
    (8, 4)
    >>> pm.ndim
    2

    >>> pm.shape
    ()
    >>> pm.ndim
    0

        '''
        return self.matrix.ndim

    @property
    def shape(self):
        '''List of the partition matrix's dimension sizes.

    Not to be confused with the sizes of the master data array's
    dimensions.

    **Examples:**

    >>> pm.ndim
    2
    >>> pm.size
    32
    >>> pm.shape
    (8, 4)

    >>> pm.ndim
    0
    >>> pm.shape
    ()

        '''
        return self.matrix.shape

    @property
    def size(self):
        '''The number of partitions in the partition matrix.

    Not to be confused with the number of elements in the master data
    array.

    **Examples:**

    >>> pm.shape
    (8, 4)
    >>> pm.size
    32

    >>> pm.shape
    ()
    >>> pm.size
    1

        '''
        return self.matrix.size

    def add_partitions(self, adimensions, master_flip, extra_boundaries, axis):
        '''Add partition boundaries.

    :Parameters:

        adimensions: `list`
            The ordered axis names of the master array.

        master_flip: `list`

        extra_boundaries: `list` of `int`
            The boundaries of the new partitions.

        axis: `str`
            The name of the axis to have the new partitions.

        '''
        def _update_p(matrix, location, master_index,
                      part, master_axis_to_position, master_flip):
            '''TODO

        :Parameters:

            matrix: numpy array of `Partition` objects

            location: `list`

            master_index: `int`

            part: `list`

            master_axis_to_position: `dict`

            master_flip: `list`

        :Returns:

            numpy array of `Partition` objects

            '''
            for partition in matrix.flat:
                partition.location = partition.location[:]
                partition.shape = partition.shape[:]

                partition.location[master_index] = location
                partition.shape[master_index] = shape

                partition.new_part(part,
                                   master_axis_to_position,
                                   master_flip)
            # --- End: for

            return matrix

        # If no extra boundaries have been provided, just return
        # without doing anything
        if not extra_boundaries:
            return

        master_index = adimensions.index(axis)
        index = self.axes.index(axis)

        # Find the position of the extra-boundaries dimension in the
        # list of master array dimensions
        extra_boundaries = extra_boundaries[:]

        # Create the master_axis_to_position dictionary required by
        # Partition.new_part
        master_axis_to_position = {}
        for i, data_axis in enumerate(adimensions):
            master_axis_to_position[data_axis] = i

        matrix = self.matrix
        shape = matrix.shape

        # Initialize the new partition matrix
        new_shape = list(shape)
        new_shape[index] += len(extra_boundaries)
        new_matrix = numpy_empty(new_shape, dtype=object)

        part = [slice(None)] * len(adimensions)
        indices = [slice(None)] * matrix.ndim
        new_indices = indices[:]
        new_indices[index] = slice(0, 1)  # 0

        # Find the first extra boundary
        x = extra_boundaries.pop(0)

        for i in range(shape[index]):
            indices[index] = slice(i, i+1)
            sub_matrix = matrix[tuple(indices)]
#            (r0, r1) = next(sub_matrix.flat).location[master_index]
            (r0, r1) = sub_matrix.item(0,).location[master_index]

# Could do better, perhaps, by assigning in blocks
            if not r0 < x < r1:
                # This new boundary is *not* within the span of this
                # sub-matrix. Therefore, just copy the sub-matrix
                # straight into the new matrix
                new_matrix[tuple(new_indices)] = sub_matrix
#                new_indices[index] += 1
                new_indices[index] = slice(
                    new_indices[index].start+1, new_indices[index].stop+1)
                continue

            # --------------------------------------------------------
            # Still here? Then this new boundary *is* within the span
            # of this sub-matrix.
            # --------------------------------------------------------

            # Find the new extent of the original partition(s)
            location = (r0, x)
            shape = x - r0
            part[master_index] = slice(0, shape)

            # Create new partition(s) in place of the original ones(s)
            # and set the location, shape and part attributes
            new_matrix[tuple(new_indices)] = _update_p(
                deepcopy(sub_matrix), location, master_index, part,
                master_axis_to_position, master_flip)
#            new_indices[index] += 1
            new_indices[index] = slice(
                new_indices[index].start+1, new_indices[index].stop+1)

            while x < r1:
                # Find the extent of the new partition(s)
                if not extra_boundaries:
                    # There are no more new boundaries, so the new
                    # partition(s) run to the end of the original
                    # partition(s) in which they lie.
                    location1 = r1
                else:
                    # There are more new boundaries, so this
                    # new partition runs either to the next
                    # new boundary or to the end of the
                    # original partition, which comes first.
                    location1 = min(extra_boundaries[0], r1)

                location = (x, location1)
                shape = location1 - x
                offset = x - r0
                part[master_index] = slice(offset, offset + shape)

                # Create the new partition(s) and set the
                # location, shape and part attributes
                new_matrix[tuple(new_indices)] = _update_p(
                    deepcopy(sub_matrix), location, master_index, part,
                    master_axis_to_position, master_flip
                )

                new_indices[index] = slice(
                    new_indices[index].start+1, new_indices[index].stop+1)
#                new_indices[index] += 1

                if not extra_boundaries:
                    # There are no more extra boundaries, so we can
                    # return now.
                    #  new_indices[index] = slice(new_indices[index],
                    #  None)
                    new_indices[index] = slice(new_indices[index].start,
                                               None)
                    indices[index] = slice(i+1, None)

                    new_matrix[tuple(new_indices)] = matrix[tuple(indices)]
                    self.matrix = new_matrix

                    return

                # Still here? Then move on to the next new boundary
                x = extra_boundaries.pop(0)
            # --- End: while
        # --- End: for

        self.matrix = new_matrix

    def copy(self):
        '''Return a deep copy.

    ``pm.copy()`` is equivalent to ``copy.deepcopy(pm)``.

    :Returns:

            The deep copy.

    **Examples:**

    >>> pm.copy()

        '''
        # ------------------------------------------------------------
        # NOTE: 15 May 2013. It is necesary to treat
        #       self.matrix.ndim==0 as a special case since there is a
        #       bug (feature?) in numpy <= v1.7 (at least):
        #       http://numpy-discussion.10968.n7.nabble.com/bug-in-deepcopy-of-rank-zero-arrays-td33705.html
        # ------------------------------------------------------------
        matrix = self.matrix

        if not matrix.ndim:
            new_matrix = _empty_matrix.copy()  # numpy_empty((), dtype=object)
            new_matrix[()] = matrix.item().copy()
            return type(self)(new_matrix, [])
        else:
            new_matrix = numpy.empty(matrix.size, dtype=object)
            new_matrix[...] = [partition.copy() for partition in matrix.flat]
            new_matrix.resize(matrix.shape)
            return type(self)(new_matrix, self.axes)

    # 0
    @_inplace_enabled
    def insert_dimension(self, axis, inplace=False):
        '''Insert a new size 1 axis in place.

    The new axis is always inserted at position 0, i.e. it becomes the
    new slowest varying axis.

    .. seealso:: `flip`, `squeezes`, `swapaxes`, `transpose`

    :Parameters:

        axis: `str`
            The internal identity of the new axis.

    :Returns:

        `PartitionMatrix`

    **Examples:**

    >>> pm.shape
    (2, 3)
    >>> pm.insert_dimension('dim2')
    >>> pm.shape
    (1, 2, 3)

        '''
        p = _inplace_enabled_define_and_cleanup(self)

        p.matrix = numpy_expand_dims(p.matrix, 0)
        p.axes = [axis] + p.axes

        return p

    def ndenumerate(self):
        '''Return an iterator yielding pairs of array indices and values.

    :Returns:

        `numpy.ndenumerate`
            An iterator over the array coordinates and values.

    **Examples:**

    >>> pm.shape
    (2, 3)
    >>> for i, partition in pm.ndenumerate():
    ...     print(i, repr(partition))
    ...
    (0, 0) <cf.data.partition.Partition object at 0x13a4490>
    (0, 1) <cf.data.partition.Partition object at 0x24a4650>
    (0, 2) <cf.data.partition.Partition object at 0x35a4590>
    (1, 0) <cf.data.partition.Partition object at 0x46a4789>
    (1, 1) <cf.data.partition.Partition object at 0x57a3456>
    (1, 2) <cf.data.partition.Partition object at 0x68a9872>

        '''
        return numpy_ndenumerate(self.matrix)

    # 0
    def partition_boundaries(self, data_axes):
        '''Return the partition boundaries for each dimension.

    :Parameters:

        data_axes: sequence

    :Returns:

        `dict`

        '''
        boundaries = {}

        matrix = self.matrix
        indices = [0] * self.ndim

        for i, axis in enumerate(self.axes):
            indices[i] = slice(None)
            j = data_axes.index(axis)

            sub_matrix = matrix[tuple(indices)]

            b = [partition.location[j][0] for partition in sub_matrix.flat]

            # Python3: can't access variables from within previous
            # list comprehension
            last_partition = sub_matrix.flat[-1]

            b.append(last_partition.location[j][1])
            boundaries[axis] = b

            indices[i] = 0
        # --- End: for

        return boundaries

    # 0
    @_inplace_enabled
    def swapaxes(self, axis0, axis1, inplace=False):
        '''Swap the positions of two axes.

    Note that this does not change the master data array.

    .. seealso:: `insert_dimension`, `flip`, `squeeze`, `transpose`

    :Parameters:

        axis0, axis1: `int`, `int`
            Select the axes to swap. Each axis is identified by its
            original integer position.

    :Returns:

        `PartitionMatrix`

    **Examples:**

    >>> pm.shape
    (2, 3, 4, 5)
    >>> pm.swapaxes(1, 2)
    >>> pm.shape
    (2, 4, 3, 5)
    >>> pm.swapaxes(1, -1)
    >>> pm.shape
    (2, 5, 3, 4)

        '''
        p = _inplace_enabled_define_and_cleanup(self)

        if axis0 != axis1:
            iaxes = list(range(p.matrix.ndim))
            iaxes[axis1], iaxes[axis0] = iaxes[axis0], iaxes[axis1]
            p.transpose(iaxes, inplace=True)

        return p

    # 0
    def set_location_map(self, data_axes, ns=None):
        '''Set the `!location` attribute of each partition of the partition
    matrix in place.

    :Parameters:

        data_axes: sequence
            The axes of the master data array.

        ns: sequence of `int`, optional

    :Returns:

        `None`

    **Examples:**

    >>> pm.set_location_map(['dim1', 'dim0'])
    >>> pm.set_location_map([])

        '''
        matrix = self.matrix

        shape = matrix.shape
        axes = self.axes

        slice_None = slice(None)

        indices = [slice_None] * matrix.ndim

        # Never update location in-place
        for partition in matrix.flat:
            partition.location = partition.location[:]

        if ns is None:
            ns = range(len(data_axes))

        for axis, n in zip(data_axes, ns):

            if axis in axes:
                # ----------------------------------------------------
                # This data array axis is also a partition matrix axis
                # ----------------------------------------------------
                m = axes.index(axis)
                start = 0
                for i in range(shape[m]):
                    indices[m] = slice(i, i+1)  # i
                    flat = matrix[tuple(indices)].flat

                    partition = next(flat)
                    stop = start + partition.shape[n]
                    location = (start, stop)

                    partition.location[n] = location

                    for partition in flat:
                        partition.location[n] = location

                    start = stop

                # --- End: for
                indices[m] = slice_None
            else:
                # ----------------------------------------------------
                # This data array axis is not a partition matrix axis
                # ----------------------------------------------------
                flat = matrix.flat
                partition = next(flat)
                location = (0, partition.shape[n])

                partition.location[n] = location

                for partition in flat:
                    partition.location[n] = location
        # --- End: for

    # 0
    @_inplace_enabled
    def squeeze(self, inplace=False):
        '''Remove all size 1 axes.

    Note that this does not change the master data array.

    .. seealso:: `insert_dimension`, `flip`, `swapaxes`, `transpose`

    :Returns:

        `PartitionMatrix`

    **Examples:**

    >>> pm.shape
    (1, 2, 1, 2)
    >>> pm.squeeze()
    >>> pm.shape
    (2, 2)

    >>> pm.shape
    (1,)
    >>> pm.squeeze()
    >>> pm.shape
    ()
    >>> pm.squeeze()
    >>> pm.shape
    ()

        '''
        p = _inplace_enabled_define_and_cleanup(self)

        matrix = p.matrix
        shape = matrix.shape

        if 1 in shape:
            p.matrix = matrix.squeeze()

            axes = p.axes
            p.axes = [axis for axis, size in zip(axes, shape) if size > 1]

        return p

    @_inplace_enabled
    def transpose(self, axes, inplace=False):
        '''Permute the partition dimensions of the partition matrix in place.

    Note that this does not change the master data array.

    .. seealso:: `insert_dimension`, `flip`, `squeeze`, `swapaxes`

    :Parameters:

        axes: sequence of `int`
            Permute the axes according to the values given.

    :Returns:

        `PartitionMatrix`

    **Examples:**

    >>> pm.ndim
    3
    >>> pm.transpose((2, 0, 1))

        '''
        p = _inplace_enabled_define_and_cleanup(self)

        matrix = p.matrix
        if list(axes) != list(range(matrix.ndim)):
            p.matrix = matrix.transpose(axes)
            p_axes = p.axes
            p.axes = [p_axes[i] for i in axes]

        return p

    # ----------------------------------------------------------------
    # Deprecated attributes and methods
    # ----------------------------------------------------------------
    def expand_dims(self, axis, inplace=False):
        '''Insert a new size 1 axis in place.

    Deprecated at version 3.0.0. Use method 'insert_dimension'
    instead.

        '''
        _DEPRECATION_ERROR_METHOD(
            self, 'expand_dims', "Use method 'insert_dimension' instead."
        )  # pragma: no cover


# --- End: class



import os

from functools import cmp_to_key

import numpy

from . import cInterface
from .extraData import ExtraDataUnpacker


class UMFileException(Exception):
    pass


class File:
    '''A class for a UM data file that gives a view of the file including
    sets of PP records combined into variables

    '''
    def __init__(self, path, byte_ordering=None, word_size=None,
                 format=None, parse=True):
        '''Open and parse a UM file.  The following optional arguments specify
    the file type. If all three are set, then this forces the file
    type; otherwise, the file type is autodetected and any of them
    that are set are ignored.

    :Parameters:

        byte_ordering: 'little_endian' or 'big_endian'

        word_size: 4 or 8

        format: 'FF' or 'PP'

    The default action is to open the file, store the file type from
    the arguments or autodetection as described above, and then parse
    the contents, giving a tree of variables and records under the
    File object. However, if "parse = False" is set, then an object is
    returned in which the last step is omitted, so only the file type
    is stored, and there are no variables under it. Such an object can
    be passed when instantiating Rec objects, and contains sufficient
    info about the file type to ensure that the get_data method of
    those Rec objects will work.

        '''
        c = cInterface.CInterface()
        self._c_interface = c
        self.path = path
        self.fd = None
        self.open_fd()
        if byte_ordering and word_size and format:
            self.format = format
            self.byte_ordering = byte_ordering
            self.word_size = word_size
        else:
            self._detect_file_type()
        self.path = path
        file_type_obj = c.create_file_type(
            self.format, self.byte_ordering, self.word_size)
        c.set_word_size(file_type_obj)
        if parse:
            info = c.parse_file(self.fd, file_type_obj)
            self.vars = info["vars"]
            self._add_back_refs()

    def open_fd(self):
        '''(Re)open the low-level file descriptor.

        '''
        if self.fd is None:
            self.fd = os.open(self.path, os.O_RDONLY)
        return self.fd

    def close_fd(self):
        '''Close the low-level file descriptor.

        '''
        if self.fd:
            os.close(self.fd)
        self.fd = None

    def _detect_file_type(self):
        c = self._c_interface
        try:
            file_type_obj = c.detect_file_type(self.fd)
        except:
            self.close_fd()
            raise IOError("File {0} has unsupported format".format(self.path))

        d = c.file_type_obj_to_dict(file_type_obj)
        self.format = d["format"]
        self.byte_ordering = d["byte_ordering"]
        self.word_size = d["word_size"]

    def _add_back_refs(self):
        '''Add file attribute to Var objects, and both file and var
        attributes to Rec objects.  The important one is the file
        attribute in the Rec object, as this is used when reading
        data.  The others are provided for extra convenience.

        '''
        for var in self.vars:
            var.file = self
            for rec in var.recs:
                rec.var = var
                rec.file = self


# --- End: class


class Var:
    '''Container for some information about variables

    '''
    def __init__(self, recs, nz, nt, supervar_index=None):
        self.recs = recs
        self.nz = nz
        self.nt = nt
        self.supervar_index = supervar_index

    @staticmethod
    def _compare(x, y):
        '''Method equivalent to the Python 2 'cmp'. Note that (x > y) - (x <
    y) is equivalent but not as performant since it would not
    short-circuit.

        '''
        if x == y:
            return 0
        elif x > y:
            return 1
        else:
            return -1

    def _compare_recs_by_extra_data(self, a, b):
        return self._compare(a.get_extra_data(), b.get_extra_data())

    def _compare_recs_by_orig_order(self, a, b):
        return self._compare(self.recs.index(a), self.recs.index(b))

    def group_records_by_extra_data(self):
        '''Returns a list of (sub)lists of records where each records within
    each sublist has matching extra data (if any), so if the whole
    variable has consistent extra data then the return value will be
    of length 1.  Within each group, the ordering of returned records
    is the same as in self.recs

        '''
        compare = self._compare_recs_by_extra_data
        recs = self.recs[:]
        n = len(recs)
        if n == 0:
            # shouldn't have a var without records, but...
            return []

#        recs.sort(compare) #python2
        recs.sort(key=cmp_to_key(compare))

        # optimise simple case - if two ends of a sorted list match,
        # the whole list matches
        if not compare(recs[0], recs[-1]):
            return [self.recs[:]]

        groups = []
        this_grp = []
        for i, rec in enumerate(recs):
            this_grp.append(rec)
            if i == n - 1 or compare(rec, recs[i + 1]):
                this_grp.sort(key=self._compare_recs_by_orig_order)
                groups.append(this_grp)
                this_grp = []
        # --- End: for

        return groups


# --- End: class


class Rec:
    '''Container for some information about records

    '''
    def __init__(self, int_hdr, real_hdr, hdr_offset, data_offset, disk_length,
                 file=None):
        '''Default instantiation, which stores the supplied headers and
    offsets.  The 'file' argument, used to set the 'file' attribute,
    does not need to be supplied, but if it is not then it will have
    to be set on the returned object (to the containing File object)
    before calling get_data() will work.  Normally this would be done
    by the calling code instantiating via File rather than directly.

        '''
        self.int_hdr = int_hdr
        self.real_hdr = real_hdr
        self.hdr_offset = hdr_offset
        self.data_offset = data_offset
        self.disk_length = disk_length
        self._extra_data = None
        if file:
            self.file = file

    @classmethod
    def from_file_and_offsets(cls, file, hdr_offset, data_offset, disk_length):
        '''Instantiate a Rec object from the file object and the header and
    data offsets. The headers are read in, and also the record object
    is ready for calling get_data().

        '''
        c = file._c_interface
        int_hdr, real_hdr = c.read_header(
            file.fd,
            hdr_offset,
            file.byte_ordering,
            file.word_size
        )
        return cls(
            int_hdr, real_hdr, hdr_offset, data_offset, disk_length, file=file)

    def read_extra_data(self):
        '''Read the extra data associated with the record

        '''
        c = self.file._c_interface
        file = self.file

        extra_data_offset, extra_data_length = (
            c.get_extra_data_offset_and_length(
                self.int_hdr,
                self.data_offset,
                self.disk_length
            )
        )

        raw_extra_data = c.read_extra_data(
            file.fd,
            extra_data_offset,
            extra_data_length,
            file.byte_ordering,
            file.word_size
        )

        edu = ExtraDataUnpacker(
            raw_extra_data,
            file.word_size,
            file.byte_ordering
        )

        return edu.get_data()

    def get_extra_data(self):
        '''Get extra data associated with the record, either by reading or
    using cached read

        '''
        if self._extra_data is None:
            self._extra_data = self.read_extra_data()

        return self._extra_data

    def get_type_and_num_words(self):
        '''Get the data type (as numpy type) and number of words as 2-tuple

        '''
        c = self.file._c_interface
        ntype, num_words = c.get_type_and_num_words(self.int_hdr)
        if ntype == 'integer':
            dtype = numpy.dtype(c.file_data_int_type)
        elif ntype == 'real':
            dtype = numpy.dtype(c.file_data_real_type)
        return dtype, num_words

    def get_data(self):
        '''Get the data array associated with the record

        '''
        c = self.file._c_interface
        file = self.file
        data_type, nwords = c.get_type_and_num_words(self.int_hdr)

        return c.read_record_data(
            file.fd,
            self.data_offset,
            self.disk_length,
            file.byte_ordering,
            file.word_size,
            self.int_hdr,
            self.real_hdr,
            data_type,
            nwords
        )


# --- End: class


if __name__ == '__main__':
    import sys

    path = sys.argv[1]
    f = File(path)
    print(f.format, f.byte_ordering, f.word_size)
    print("num variables: %s" % len(f.vars))
    for varno, var in enumerate(f.vars):
        print()
        print("var %s: nz = %s, nt = %s" % (varno, var.nz, var.nt))
        for recno, rec in enumerate(var.recs):
            print("var %s record %s" % (varno, recno))
            print("hdr offset: %s" % rec.hdr_offset)
            print("data offset: %s" % rec.data_offset)
            print("disk length: %s" % rec.disk_length)
            print("int hdr: %s" % rec.int_hdr)
            print("real hdr: %s" % rec.real_hdr)
            print("data: %s" % rec.get_data())
            print("extra_data: %s" % rec.get_extra_data())
            print("type %s, num words: %s" % rec.get_type_and_num_words())
            # if recno == 1:
            #     rec._extra_data['y'] += .01
            #     print("massaged_extra_data: %s" % rec.get_extra_data())
            print("-----------------------")

        print("all records", var.recs)
        print("records grouped by extra data ",
              var.group_records_by_extra_data())
        print("===============================")

    f.close_fd()

    # also read a record using saved metadata
    if f.vars:

        format = f.format
        byte_ordering = f.byte_ordering
        word_size = f.word_size
        myrec = f.vars[0].recs[0]
        hdr_offset = myrec.hdr_offset
        data_offset = myrec.data_offset
        disk_length = myrec.disk_length

        del(f)

        fnew = File(
            path,
            format=format,
            byte_ordering=byte_ordering,
            word_size=word_size,
            parse=False
        )

        rnew = Rec.from_file_and_offsets(
            fnew, hdr_offset, data_offset, disk_length)
        print("record read using saved file type and offsets:")
        print("int hdr: %s" % rnew.int_hdr)
        print("real hdr: %s" % rnew.real_hdr)
        print("data: %s" % rnew.get_data())
        print("extra data: %s" % rnew.get_extra_data())
        print("nx = %s" % rnew.int_hdr[18])
        print("ny = %s" % rnew.int_hdr[17])
        rdata = open("recdata0.txt", "w")
        for value in rnew.get_data():
            rdata.write("%s\n" % value)
        rdata.close()



from numpy.ma import masked_where as numpy_ma_masked_where

from ..constants import _file_to_fh
from ..functions import (open_files_threshold_exceeded,
                         close_one_file,
                         parse_indices,
                         get_subspace)

from ..data.filearray import FileArray

from .functions import _open_um_file, _close_um_file

from .umread.umfile import Rec

_filename_to_file = _file_to_fh.setdefault('UM', {})


# ====================================================================
#
# UMFileArray object
#
# ====================================================================

class UMFileArray(FileArray):
    '''A sub-array stored in a PP or UM fields file.

**Initialization**

:Parameters:

    file : str
        The file name in normalized, absolute form.

    dtype : numpy.dtype
        The data type of the data array on disk.

    ndim : int
        The number of dimensions in the unpacked data array.

    shape : tuple
        The shape of the unpacked data array.

    size : int
        The number of elements in the unpacked data array.

    header_offset : int
        The start position in the file of the header.

    data_offset : int
        The start position in the file of the data array.

    disk_length : int
        The number of words on disk for the data array, usually
        LBLREC-LBEXT. If set to 0 then `!size` is used.

    fmt: `str`, optional

    word_size: `int`, optional

    byte_ordering: `str`, optional

:Examples:

>>> a = UMFileArray(file='file.pp', header_offset=3156, data_offset=3420,
...                 dtype=numpy.dtype('float32'), shape=(30, 24),
...                 size=720, ndim=2, disk_length=0)

>>> a = UMFileArray(file='packed_file.pp', header_offset=3156,
...                 data_offset=3420, dtype=numpy.dtype('float32'),
...                 shape=(30, 24), size=720, ndim=2, disk_length=423)

    '''
    def __getitem__(self, indices):
        '''

Implement indexing

x.__getitem__(indices) <==> x[indices]

Returns a numpy array.

'''
        f = self.open()

        rec = Rec.from_file_and_offsets(f,
                                        self.header_offset,
                                        self.data_offset,
                                        self.disk_length)

        int_hdr = rec.int_hdr
        real_hdr = rec.real_hdr

        array = rec.get_data().reshape(int_hdr.item(17,), int_hdr.item(18,))

        if indices is not Ellipsis:
            indices = parse_indices(array.shape, indices)
            array = get_subspace(array, indices)

        LBUSER2 = int_hdr.item(38,)

        if LBUSER2 == 3:
            # Return the numpy array now if it is a boolean array
            return array.astype(bool)

        integer_array = LBUSER2 == 2

        # ------------------------------------------------------------
        # Convert to a masked array
        # ------------------------------------------------------------
        # Set the fill_value from BMDI
        fill_value = real_hdr.item(17,)
        if fill_value != -1.0e30:
            # -1.0e30 is the flag for no missing data
            if integer_array:
                # The fill_value must be of the same type as the data
                # values
                fill_value = int(fill_value)

            # Mask any missing values
            mask = (array == fill_value)
            if mask.any():
                array = numpy_ma_masked_where(mask, array, copy=False)
        # --- End: if

        # ------------------------------------------------------------
        # Unpack the array using the scale_factor and add_offset, if
        # either is available
        # ------------------------------------------------------------
        # Treat BMKS as a scale_factor if it is neither 0 nor 1
        scale_factor = real_hdr.item(18,)
        if scale_factor != 1.0 and scale_factor != 0.0:
            if integer_array:
                scale_factor = int(scale_factor)
            array *= scale_factor

        # Treat BDATUM as an add_offset if it is not 0
        add_offset = real_hdr.item(4,)
        if add_offset != 0.0:
            if integer_array:
                add_offset = int(add_offset)
            array += add_offset

        # Return the numpy array
        return array
    # --- End: def

    def __str__(self):
        '''

x.__str__() <==> str(x)

'''
        return "%s%s in %s" % (self.header_offset, self.shape, self.file)
    # --- End: def

    @property
    def file_pointer(self):
        '''
'''
        return (self.file, self.header_offset)
    # --- End: def

    def close(self):
        '''

Close the file containing the data array.

If the file is not open then no action is taken.

:Returns:

    None

:Examples:

>>> f.close()

'''
        _close_um_file(self.file)
    # --- End: def

    def open(self):
        '''

Open the file containing the data array.

:Returns:

    out : um.umread.umfile.File

:Examples:

>>> f.open()

'''
        return _open_um_file(
            self.file, fmt=getattr(self, 'fmt', None),
            word_size=getattr(self, 'word_size', None),
            byte_ordering=getattr(self, 'byte_ordering', None)
        )
    # --- End: def

# --- End: class



import datetime
import os
import time
import unittest

import numpy

import cf

class CellMeasureTest(unittest.TestCase):
    def setUp(self):
        self.filename = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                     'test_file.nc')

    def test_CellMeasure__repr__str__dump(self):
        f = cf.read(self.filename)[0]
        x = f.cell_measures('measure:area').value()

        _ = repr(x)
        _ = str(x)
        _ = x.dump(display=False)

        self.assertTrue(x.ismeasure)


    def test_CellMeasure_measure(self):
        f = cf.read(self.filename)[0]
        x = f.cell_measures('measure:area').value()

        self.assertTrue(x.measure == 'area')
        del x.measure
        self.assertTrue(getattr(x, 'measure', None) is None)
        x.measure = 'qwerty'
        self.assertTrue(x.measure == 'qwerty')


    def test_CellMeasure_identity(self):
        f = cf.read(self.filename)[0]
        x = f.cell_measures('measure:area').value()

        self.assertTrue(x.identity() == 'measure:area')
        del x.measure
        self.assertTrue(x.identity() == 'ncvar%cell_measure', x.identity())
        x.nc_del_variable()
        self.assertTrue(x.identity() == '')


# --- End: class

if __name__ == "__main__":
    print('Run date:', datetime.datetime.now())
    cf.environment()
    print()
    unittest.main(verbosity=2)
